# -*- coding: utf-8 -*-
# pylint: disable=too-many-lines
"""
Perform transformations on Dimensions
=====================================

Provide functions and classes to build a Space which an algorithm can operate on.

"""
import functools
import itertools
from abc import ABCMeta, abstractmethod

import numpy

from orion.algo.space import Categorical, Dimension, Fidelity, Integer, Real, Space

NON_LINEAR = ["loguniform", "reciprocal"]


# pylint: disable=unused-argument
@functools.singledispatch
def build_transform(dim, type_requirement, dist_requirement):
    """Base transformation factory

    Parameters
    ----------
    dim: `orion.algo.space.Dimension`
        A dimension object which may need transformations to match provided requirements.
    type_requirement: str, None
        String defining the requirement of the algorithm. It can be one of the following
        - 'real', the dim should be transformed so type is `orion.algo.space.Real`
        - 'integer', the dim should be transformed so type is `orion.algo.space.Integer`
        - 'numerical', the dim should be transformed so type is either `orion.algo.space.Integer` or
        `orion.algo.space.Real`
        - None, no requirement
    dist_requirement: str, None
        String defining the distribution requirement of the algorithm.
        - 'linear', any dimension with logarithmic prior while be linearized
        - None, no requirement

    """
    return []


@build_transform.register(Categorical)
def _(dim, type_requirement, dist_requirement):
    transformers = []
    if type_requirement == "real":
        transformers.extend(
            [Enumerate(dim.categories), OneHotEncode(len(dim.categories))]
        )
    elif type_requirement in ["integer", "numerical"]:
        transformers.append(Enumerate(dim.categories))

    return transformers


@build_transform.register(Fidelity)
def _(dim, type_requirement, dist_requirement):
    return []


@build_transform.register(Integer)
def _(dim, type_requirement, dist_requirement):
    transformers = []
    if dist_requirement == "linear" and dim.prior_name[4:] in NON_LINEAR:
        transformers.extend([Reverse(Quantize()), Linearize()])
        # Turn back to integer because Linearize outputs real
        if type_requirement != "real":
            transformers.extend([Quantize()])
    elif type_requirement == "real":
        transformers.append(Reverse(Quantize()))

    return transformers


@build_transform.register(Real)
def _(dim, type_requirement, dist_requirement):
    transformers = []
    if dim.precision is not None:
        transformers.append(Precision(dim.precision))

    if dist_requirement == "linear" and dim.prior_name in NON_LINEAR:
        transformers.append(Linearize())
    elif type_requirement == "integer":
        transformers.append(Quantize())

    return transformers


def transform(original_space, type_requirement, dist_requirement):
    """Build a transformed space"""
    space = TransformedSpace(original_space)
    for dim in original_space.values():
        transformers = build_transform(dim, type_requirement, dist_requirement)
        space.register(
            TransformedDimension(
                transformer=Compose(transformers, dim.type), original_dimension=dim
            )
        )

    return space


def reshape(space, shape_requirement):
    """Build a reshaped space"""
    if shape_requirement is None:
        return space

    # We assume shape_requirement == 'flattened'

    reshaped_space = ReshapedSpace(space)

    for dim_index, dim in enumerate(space.values()):
        if not dim.shape or numpy.prod(dim.shape) == 1:
            reshaped_space.register(
                ReshapedDimension(
                    transformer=Identity(dim.type),
                    original_dimension=dim,
                    index=dim_index,
                )
            )
        else:
            for index in itertools.product(*map(range, dim.shape)):
                key = f'{dim.name}[{",".join(map(str, index))}]'
                reshaped_space.register(
                    ReshapedDimension(
                        transformer=View(dim.shape, index, dim.type),
                        original_dimension=dim,
                        name=key,
                        index=dim_index,
                    )
                )

    return reshaped_space


def build_required_space(
    original_space, type_requirement=None, shape_requirement=None, dist_requirement=None
):
    """Build a :class:`orion.algo.space.Space` object which agrees to the `requirements` imposed
    by the desired optimization algorithm.

    It uses appropriate cascade of `Transformer` objects per `orion.algo.space.Dimension`
    contained in `original_space`. `ReshapedTransformer` objects are used above
    the `Transformer` if the optimizatios algorithm requires flattened dimensions.

    Parameters
    ----------
    original_space : `orion.algo.space.Space`
        Original problem's definition of parameter space given by the user to Oríon.
    type_requirement: str, None
        String defining the requirement of the algorithm. It can be one of the following
        - 'real', the dim should be transformed so type is `orion.algo.space.Real`
        - 'integer', the dim should be transformed so type is `orion.algo.space.Integer`
        - 'numerical', the dim should be transformed so type is either `orion.algo.space.Integer` or
        `orion.algo.space.Real`
        - None, no requirement
    shape_requirement: str, None
        String defining the shape requirement of the algorithm.
        - 'flattened', any dimension with shape > 1 will be flattened
        - None, no requirement
    dist_requirement: str, None
        String defining the distribution requirement of the algorithm.
        - 'linear', any dimension with logarithmic prior while be linearized
        - None, no requirement

    """
    space = transform(original_space, type_requirement, dist_requirement)
    space = reshape(space, shape_requirement)

    return space


class Transformer(object, metaclass=ABCMeta):
    """Define an (injective) function and its inverse. Base transformation class.

    Attributes
    ----------
    target_type: str
        Defines the type of the target space of the forward function.
        It can provide one of the values: ``['real', 'integer', 'categorical']``.
    domain_type: str
        Is similar to ``target_type`` but it refers to the domain.
        If it is ``None``, then it can receive inputs of any type.

    """

    domain_type = None
    target_type = None

    @abstractmethod
    def transform(self, point):
        """Transform a point from domain dimension to the target dimension."""
        pass

    @abstractmethod
    def reverse(self, transformed_point, index=None):
        """Reverse transform a point from target dimension to the domain dimension."""
        pass

    # pylint:disable=no-self-use
    def infer_target_shape(self, shape):
        """Return the shape of the dimension after transformation."""
        return shape

    def repr_format(self, what):
        """Format a string for calling ``__repr__`` in `TransformedDimension`."""
        return "{}({})".format(self.__class__.__name__, what)

    def _get_hashable_members(self):
        return (self.__class__.__name__, self.domain_type, self.target_type)

    # pylint:disable=protected-access
    def __eq__(self, other):
        """Return True if other is the same transformed dimension as self"""
        if not isinstance(other, Transformer):
            return False
        return self._get_hashable_members() == other._get_hashable_members()


class Identity(Transformer):
    """Implement an identity transformation. Everything as it is."""

    def __init__(self, domain_type=None):
        self._domain_type = domain_type

    @property
    def first(self):
        """Signals to ReshapedSpace whether this dimension should be used for `reverse`"""
        return True

    def transform(self, point):
        """Return `point` as it is."""
        return point

    # pylint:disable=unused-argument
    def reverse(self, transformed_point, index=None):
        """Return `transformed_point` as it is."""
        if index is not None:
            return transformed_point[index]
        return transformed_point

    def repr_format(self, what):
        """Format a string for calling ``__repr__`` in `TransformedDimension`."""
        return what

    @property
    def domain_type(self):
        """Return declared domain type on initialization."""
        return self._domain_type

    @property
    def target_type(self):
        """Return domain type as this will be the target in a identity transformation."""
        return self.domain_type


class Compose(Transformer):
    """Initialize composite transformer with a list of `Transformer` objects
    and domain type on which it will be applied.
    """

    def __init__(self, transformers, base_domain_type=None):
        try:
            self.apply = transformers[-1]
        except IndexError:
            self.apply = Identity()
        if len(transformers) > 1:
            self.composition = Compose(transformers[:-1], base_domain_type)
        else:
            self.composition = Identity(base_domain_type)
        assert (
            self.apply.domain_type is None
            or self.composition.target_type == self.apply.domain_type
        )

    def transform(self, point):
        """Apply transformers in the increasing order of the `transformers` list."""
        point = self.composition.transform(point)
        return self.apply.transform(point)

    # pylint:disable=unused-argument
    def reverse(self, transformed_point, index=None):
        """Reverse transformation by reversing in the opposite order of the `transformers` list."""
        transformed_point = self.apply.reverse(transformed_point)
        return self.composition.reverse(transformed_point)

    def interval(self, alpha=1.0):
        """Return interval of composed transformation."""
        if hasattr(self.apply, "interval"):
            return self.apply.interval(alpha)

        return None

    def infer_target_shape(self, shape):
        """Return the shape of the dimension after transformation."""
        shape = self.composition.infer_target_shape(shape)
        return self.apply.infer_target_shape(shape)

    def repr_format(self, what):
        """Format a string for calling ``__repr__`` in `TransformedDimension`."""
        return self.apply.repr_format(self.composition.repr_format(what))

    @property
    def domain_type(self):
        """Return base domain type."""
        return self.composition.domain_type

    @property
    def target_type(self):
        """Infer type of the tranformation target."""
        type_before = self.composition.target_type
        type_after = self.apply.target_type
        return type_after if type_after else type_before

    # pylint:disable=protected-access
    def _get_hashable_members(self):
        return (
            (self.__class__.__name__,)
            + self.apply._get_hashable_members()
            + self.composition._get_hashable_members()
        )


class Reverse(Transformer):
    """Apply the reverse transformation that another one would do."""

    def __init__(self, transformer: Transformer):
        assert not isinstance(
            transformer, OneHotEncode
        ), "real to categorical is pointless"
        self.transformer = transformer

    def transform(self, point):
        """Use `reserve` of composed `transformer`."""
        return self.transformer.reverse(point)

    # pylint:disable=unused-argument
    def reverse(self, transformed_point, index=None):
        """Use `transform` of composed `transformer`."""
        return self.transformer.transform(transformed_point)

    def repr_format(self, what):
        """Format a string for calling ``__repr__`` in `TransformedDimension`."""
        return "{}{}".format(
            self.__class__.__name__, self.transformer.repr_format(what)
        )

    @property
    def target_type(self):
        """Return `domain_type` of composed `transformer`."""
        return self.transformer.domain_type

    @property
    def domain_type(self):
        """Return `target_type` of composed `transformer`."""
        return self.transformer.target_type


class Precision(Transformer):
    """Round real numbers to requested precision."""

    domain_type = "real"
    target_type = "real"

    def __init__(self, precision=4):
        self.precision = precision

    def transform(self, point):
        """Round `point` to the requested precision, as numpy arrays."""
        # numpy.format_float_scientific precision starts at 0
        if isinstance(point, (list, tuple)) or (
            isinstance(point, numpy.ndarray) and point.shape
        ):
            format_float = numpy.vectorize(
                lambda x: numpy.format_float_scientific(x, precision=self.precision - 1)
            )
            point = format_float(point)
            to_float = numpy.vectorize(float)
            point = to_float(point)
        else:
            point = float(
                numpy.format_float_scientific(point, precision=self.precision - 1)
            )

        return numpy.asarray(point)

    # pylint:disable=unused-argument
    def reverse(self, transformed_point, index=None):
        """Cast `transformed_point` to floats, as numpy arrays."""
        return self.transform(transformed_point)

    def repr_format(self, what):
        """Format a string for calling ``__repr__`` in `TransformedDimension`."""
        return "{}({}, {})".format(self.__class__.__name__, self.precision, what)


class Quantize(Transformer):
    """Transform real numbers to integers, violating injection."""

    domain_type = "real"
    target_type = "integer"

    def transform(self, point):
        """Cast `point` to the floor and then to integers, as numpy arrays."""
        quantized = numpy.floor(numpy.asarray(point)).astype(int)

        if numpy.any(numpy.isinf(point)):
            isinf = int(numpy.isinf(point))
            quantized = (
                isinf * (quantized - 1) * int(numpy.sign(point))
                + (1 - isinf) * (quantized - 1)
            ).astype(int)

        return quantized

    # pylint:disable=unused-argument
    def reverse(self, transformed_point, index=None):
        """Cast `transformed_point` to floats, as numpy arrays."""
        return numpy.asarray(transformed_point).astype(float)


class Enumerate(Transformer):
    """Enumerate categories.

    Effectively transform from a list of objects to a range of integers.
    """

    domain_type = "categorical"
    target_type = "integer"

    def __init__(self, categories):
        self.categories = categories
        map_dict = {cat: i for i, cat in enumerate(categories)}
        self._map = numpy.vectorize(lambda x: map_dict[x], otypes="i")
        self._imap = numpy.vectorize(lambda x: categories[x], otypes=[numpy.object])

    def __deepcopy__(self, memo):
        """Make a deepcopy"""
        return type(self)(self.categories)

    def transform(self, point):
        """Return integers corresponding uniquely to the categories in `point`."""
        return self._map(point)

    # pylint:disable=unused-argument
    def reverse(self, transformed_point, index=None):
        """Return categories corresponding to their positions inside `transformed_point`."""
        return self._imap(transformed_point)

    # pylint:disable=unused-argument
    def interval(self, alpha=1.0):
        """Return the interval for the enumerated choices."""
        return (0, len(self.categories) - 1)


class OneHotEncode(Transformer):
    """Encode categories to a 1-hot integer space representation."""

    domain_type = "integer"
    target_type = "real"

    def __init__(self, bound: int):
        self.num_cats = bound

    def transform(self, point):
        """Match a `point` containing integers to real vector representations of them.

        If the upper bound of integers supported by an instance of `OneHotEncode`
        is less or equal to 2, then cast them to floats.

        .. note:: This transformation possibly appends one more tensor dimension to `point`.
        """
        point_ = numpy.asarray(point)
        assert (
            numpy.all(point_ < self.num_cats)
            and numpy.all(point_ >= 0)
            and numpy.all(point_ % 1 == 0)
        )

        if self.num_cats <= 2:
            return numpy.asarray(point_, dtype=float)

        hot = numpy.zeros(self.infer_target_shape(point_.shape))
        grid = numpy.meshgrid(
            *[numpy.arange(dim) for dim in point_.shape], indexing="ij"
        )
        hot[grid + [point_]] = 1
        return hot

    # pylint:disable=unused-argument
    def reverse(self, transformed_point, index=None):
        """Match real vector representations to integers using an argmax function.

        If the number of dimensions is exactly 2, then use 0.5 as a decision boundary,
        and convert representation to integers 0 or 1.

        If the number of dimensions is exactly 1, then return zeros.

        .. note:: This reverse transformation possibly removes the last tensor dimension
           from `transformed_point`.
        """
        point_ = numpy.asarray(transformed_point)
        if self.num_cats == 2:
            return (point_ > 0.5).astype(int)
        elif self.num_cats == 1:
            return numpy.zeros_like(point_, dtype=int)

        assert point_.shape[-1] == self.num_cats
        return point_.argmax(axis=-1)

    # pylint:disable=unused-argument
    def interval(self, alpha=1.0):
        """Return the interval for the one-hot encoding in proper shape."""
        if self.num_cats == 2:
            return 0, 1
        else:
            low = numpy.zeros(self.num_cats)
            high = numpy.ones(self.num_cats)

            return low, high

    def infer_target_shape(self, shape):
        """Infer that transformed points will have one more tensor dimension,
        if the number of supported integers to transform is larger than 2.
        """
        return tuple(list(shape) + [self.num_cats]) if self.num_cats > 2 else shape

    def _get_hashable_members(self):
        return super(OneHotEncode, self)._get_hashable_members() + (self.num_cats,)


class Linearize(Transformer):
    """Transform real numbers from loguniform to linear."""

    domain_type = "real"
    target_type = "real"

    def transform(self, point):
        """Linearize logarithmic distribution."""
        return numpy.log(numpy.asarray(point))

    # pylint:disable=unused-argument
    def reverse(self, transformed_point, index=None):
        """Turn linear distribution to logarithmic distribution."""
        return numpy.exp(numpy.asarray(transformed_point))


class View(Transformer):
    """Look-up single index in a dimensions with shape > 1"""

    def __init__(self, shape, index, domain_type=None):
        self.shape = shape
        self.index = index
        self._domain_type = domain_type

    @property
    def first(self):
        """Signals to ReshapedSpace whether this dimension should be used for `reverse`"""
        return sum(self.index) == 0

    def transform(self, point):
        """Only return one element of the group"""
        return point[self.index]

    def reverse(self, transformed_point, index=None):
        """Only return packend point if view of first element, otherwise drop."""
        subset = transformed_point[index : index + numpy.prod(self.shape)]
        return numpy.array(subset).reshape(self.shape)

    def interval(self, interval):
        """Return corresponding view from interval"""
        return (interval[0][self.index], interval[1][self.index])

    @property
    def domain_type(self):
        """Return declared domain type on initialization."""
        return self._domain_type

    @property
    def target_type(self):
        """Return domain type as this will be the target in flatten transformation."""
        return self.domain_type

    def repr_format(self, what):
        """Format a string for calling ``__repr__`` in `TransformedDimension`."""
        return "{}(shape={}, index={}, {})".format(
            self.__class__.__name__, self.shape, self.index, what
        )


class TransformedDimension(object):
    """Duck-type :class:`orion.algo.space.Dimension` to mimic its functionality,
    while transform automatically and appropriately an underlying
    :class:`orion.algo.space.Dimension` object according to a `Transformer` object.
    """

    NO_DEFAULT_VALUE = Dimension.NO_DEFAULT_VALUE

    def __init__(self, transformer, original_dimension):
        self.original_dimension = original_dimension
        self.transformer = transformer

    def transform(self, point):
        """Expose `Transformer.transform` interface from underlying instance."""
        return self.transformer.transform(point)

    # pylint:disable=unused-argument
    def reverse(self, transformed_point, index=None):
        """Expose `Transformer.reverse` interface from underlying instance."""
        return self.transformer.reverse(transformed_point)

    def interval(self, alpha=1.0):
        """Map the interval bounds to the transformed ones."""
        if hasattr(self.transformer, "interval"):
            interval = self.transformer.interval()
            if interval:
                return interval
        if self.original_dimension.type == "categorical":
            return self.original_dimension.categories

        low, high = self.original_dimension.interval(alpha)

        return self.transform(low), self.transform(high)

    def __contains__(self, point):
        """Reverse transform and ask the original dimension if it is a possible
        sample.
        """
        try:
            orig_point = self.reverse(point)
        except AssertionError:
            return False
        return orig_point in self.original_dimension

    def __repr__(self):
        """Represent the object as a string."""
        return self.transformer.repr_format(repr(self.original_dimension))

    # pylint:disable=protected-access
    def __eq__(self, other):
        """Return True if other is the same transformed dimension as self"""
        if not (hasattr(other, "transformer") and hasattr(other, "original_dimension")):
            return False

        return (
            self.transformer == other.transformer
            and self.original_dimension == other.original_dimension
        )

    def __hash__(self):
        """Hash of the transformed dimension"""
        return hash(self._get_hashable_members())

    # pylint:disable=protected-access
    def _get_hashable_members(self):
        """Hashable members of transformation and original dimension"""
        return (
            self.transformer._get_hashable_members()
            + self.original_dimension._get_hashable_members()
        )

    def validate(self):
        """Validate original_dimension"""
        self.original_dimension.validate()

    @property
    def name(self):
        """Do not change the name of the original dimension."""
        return self.original_dimension.name

    @property
    def type(self):
        """Ask transformer which is its target class."""
        type_ = self.transformer.target_type
        return type_ if type_ != "invariant" else self.original_dimension.type

    @property
    def prior_name(self):
        """Do not change the prior name of the original dimension."""
        return self.original_dimension.prior_name

    @property
    def shape(self):
        """Wrap original shape with transformer, because it may have changed."""
        return self.transformer.infer_target_shape(self.original_dimension.shape)

    @property
    def cardinality(self):
        """Wrap original :class:`orion.algo.space.Dimension` capacity"""
        if self.type == "real":
            return Real.get_cardinality(self.shape, self.interval())
        elif self.type == "integer":
            return Integer.get_cardinality(self.shape, self.interval())
        elif self.type == "categorical":
            return Categorical.get_cardinality(self.shape, self.interval())
        elif self.type == "fidelity":
            return Fidelity.get_cardinality(self.shape, self.interval())
        else:
            raise RuntimeError(f"No cardinality can be computed for type `{self.type}`")


class ReshapedDimension(TransformedDimension):
    """Duck-type :class:`orion.algo.space.Dimension` to mimic its functionality."""

    def __init__(self, transformer, original_dimension, index, name=None):
        super(ReshapedDimension, self).__init__(transformer, original_dimension)
        if name is None:
            name = original_dimension.name
        self._name = name
        self.index = index

    @property
    def first(self):
        """Signals to ReshapedSpace whether this dimension should be used for `reverse`"""
        return self.transformer.first

    def transform(self, point):
        """Expose `Transformer.transform` interface from underlying instance."""
        return self.transformer.transform(point[self.index])

    def reverse(self, transformed_point, index=None):
        """Expose `Transformer.reverse` interface from underlying instance."""
        return self.transformer.reverse(transformed_point, index)

    def interval(self, alpha=1.0):
        """Map the interval bounds to the transformed ones."""
        interval = self.original_dimension.interval(alpha)
        if hasattr(interval[0], "shape") and numpy.prod(interval[0].shape) > 1:
            return self.transformer.interval(interval)

        return interval

    @property
    def cardinality(self):
        """Compute cardinality"""
        cardinality = super(ReshapedDimension, self).cardinality
        if isinstance(self.transformer, View):
            cardinality /= numpy.prod(self.transformer.shape)

        return cardinality

    def cast(self, point):
        """Cast a point according to original_dimension and then transform it"""
        return self.original_dimension.cast(point)

    @property
    def shape(self):
        """Shape is fixed to ()."""
        return ()

    @property
    def name(self):
        """Name of the view"""
        return self._name


class TransformedSpace(Space):
    """Wrap the :class:`orion.algo.space.Space` to support transformation methods.

    Parameter
    ---------
    space: `orion.algo.space.Space`
       Original problem's definition of parameter space.

    """

    contains = TransformedDimension

    def __init__(self, space, *args, **kwargs):
        super(TransformedSpace, self).__init__(*args, **kwargs)
        self._original_space = space

    def transform(self, point):
        """Transform a point that was in the original space to be in this one."""
        return tuple([dim.transform(point[i]) for i, dim in enumerate(self.values())])

    def reverse(self, transformed_point):
        """Reverses transformation so that a point from this `TransformedSpace`
        to be in the original one.
        """
        return tuple(
            [dim.reverse(transformed_point[i]) for i, dim in enumerate(self.values())]
        )

    def sample(self, n_samples=1, seed=None):
        """Sample from the original dimension and forward transform them."""
        points = self._original_space.sample(n_samples=n_samples, seed=seed)
        return [self.transform(point) for point in points]


class ReshapedSpace(Space):
    """Wrap the `TransformedSpace` to support reshape methods.

    Parameter
    ---------
    space: `orion.core.worker.TransformedSpace`
       Transformed version of the orinigal problem's definition of parameter space.

    """

    contains = ReshapedDimension

    def __init__(self, original_space, *args, **kwargs):
        super(ReshapedSpace, self).__init__(*args, **kwargs)
        self._original_space = original_space

    @property
    def original(self):
        """Original space without reshape or transformations"""
        return self._original_space

    def transform(self, point):
        """Transform a point that was in the original space to be in this one."""
        return self.reshape(self.original.transform(point))

    def reverse(self, transformed_point):
        """Reverses transformation so that a point from this `ReshapedSpace` to be in the original
        one.
        """
        return self.original.reverse(self.restore_shape(transformed_point))

    def reshape(self, point):
        """Reshape the point"""
        return tuple([dim.transform(point) for dim in self.values()])

    def restore_shape(self, transformed_point):
        """Restore shape"""
        point = []
        for index, dim in enumerate(self.values()):
            if dim.first:
                point.append(dim.reverse(transformed_point, index))

        return point

    def sample(self, n_samples=1, seed=None):
        """Sample from the original dimension and forward transform them."""
        points = self.original.sample(n_samples=n_samples, seed=seed)
        return [self.reshape(point) for point in points]

    def __contains__(self, value):
        """Check whether `value` is within the bounds of the space.
        Or check if a name for a dimension is registered in this space.

        Parameters
        ----------
        value: list
            List of values associated with the dimensions contained or a string indicating a
            dimension's name.

        """
        if isinstance(value, str):
            return super(ReshapedSpace, self).__contains__(value)

        try:
            len(value)
        except TypeError as exc:
            raise TypeError(
                "Can check only for dimension names or "
                "for tuples with parameter values."
            ) from exc

        if not self:
            return False

        return self.restore_shape(value) in self.original

    @property
    def cardinality(self):
        """Reshape does not affect cardinality"""
        return self.original.cardinality



"""
====================
Parallel Coordinates
====================

.. hint:: 

   Conveys a dense overview of the trial objectives in a multi-dimensional space.
   Helps identifying trends of best or worst hyperparameter values.

The parallel coordinates plot decomposes a search space of `n` dimensions into `n`
axis so that the entire space can be visualized simultaneously. Each dimension
is represented as a vertical axis and trials are represented as lines crossing each 
axis at the corresponding value of the hyperparameters. There is no obvious optimal ordering
for the vertical axis, and you will often find that changing the order helps better understanding
the data. Additionaly, the lines are plotted with graded colors based on the objective. The
gradation is shown in a color bar on the right of the plot. Note that the objectives are added
as the last axis is the plot as well.

.. autofunction:: orion.plotting.base.parallel_coordinates
    :noindex:

The parallel coordinates plot can be executed directly from the ``experiment`` with
``plot.parallel_coordinates()`` as shown in the example below.

"""
from orion.client import get_experiment

# Specify the database where the experiments are stored. We use a local PickleDB here.
storage = dict(type="legacy", database=dict(type="pickleddb", host="../db.pkl"))

# Load the data for the specified experiment
experiment = get_experiment("2-dim-exp", storage=storage)
fig = experiment.plot.parallel_coordinates()
fig

#%%
# In this basic example the parallel coordinates plot is marginally useful as there are only
# 2 dimensions. It is possible however to identify the best performing values of ``dropout`` and
# ``learning_rate``. The GIF below demonstrates how to select subsets of the
# axis to highlight the trials that corresponds to the best objectives.
#
# .. image:: ../_static/parallel_coordinates_select.gif
#     :width: 600
#     :align: center
#
# .. note::
#
#     Hover is not supported by plotly at the moment.
#     Feature request can be tracked `here <https://github.com/plotly/plotly.js/issues/3012>`_.

#%%
# Lets now load the results from tutorial
# :ref:`sphx_glr_auto_tutorials_code_2_hyperband_checkpoint.py` for an example with a larger search
# space.

# Load the data for the specified experiment
experiment = get_experiment("hyperband-cifar10", storage=storage)
fig = experiment.plot.parallel_coordinates()
fig

#%%
# As you can see, the large number of trials until trained for a few epochs is cluttering the entire
# plot. You can first select the trials with 120 epochs to clear the plot. Once that is done,
# We can see that gamma and momentum had limited influence. Good trials can be found
# for almost any values of gamma and momentum. On the other hand, learning rate
# and weight decay are clearly more optimal in lower values. You can try re-ordering the columns as
# shown in the animation below to see the connections between one hyperparameter and the objective.

#%%
# .. image:: ../_static/parallel_coordinates_reorder.gif
#     :width: 600
#     :align: center
#

#%%
# We can also select a subset of hyperparameters to help with the visualization.

# Load the data for the specified experiment
fig = experiment.plot.parallel_coordinates(
    order=["epochs", "learning_rate", "weight_decay"]
)
fig


#%%
#
# Special cases
# -------------
#
# Logarithmic scale
# ~~~~~~~~~~~~~~~~~
#
# .. note::
#
#     Logarithmic scales are not supported yet. Contributions are welcome. :)
#     See `issue <https://github.com/Epistimio/orion/issues/555>`_.
#
# Dimension with shape
# ~~~~~~~~~~~~~~~~~~~~
#
# If some dimensions have a :ref:`search-space-shape` larger than 1, they will be flattened so that
# each subdimension can be represented in the parallel coordinates plot.

# Load the data for the specified experiment
experiment = get_experiment("2-dim-shape-exp", storage=storage)
fig = experiment.plot.parallel_coordinates()
fig

#%%
# In the example above, the dimension ``learning_rate~loguniform(1e-5, 1e-2, shape=3)``
# is flattened and represented with ``learning_rate[i]``. If the shape would be or more dimensions
# (ex: ``(3, 2)``), the indices would be ``learning_rate[i,j]`` with i=0..2 and j=0..1.

#%%
# The flattened hyperparameters can be fully selected with ``params=['<name>']``.
#

experiment.plot.parallel_coordinates(order=["/learning_rate"])

#%%
# Or a subset of the flattened hyperparameters can be selected with ``params=['<name>[index]']``.
#

experiment.plot.parallel_coordinates(order=["/learning_rate[0]", "/learning_rate[1]"])


#%%
# Categorical dimension
# ~~~~~~~~~~~~~~~~~~~~~
#
# Parallel coordinates plots can also render categorical dimensions, in which case the
# categories are shown in an arbitrary order on the axis.

# Load the data for the specified experiment
experiment = get_experiment("3-dim-cat-shape-exp", storage=storage)
fig = experiment.plot.parallel_coordinates()
fig

#%%
# Finally we save the image to serve as a thumbnail for this example. See
# the guide
# :ref:`How to save <sphx_glr_auto_examples_how-tos_code_2_how_to_save.py>`
# for more information on image saving.

fig.write_image("../../docs/src/_static/pcp_thumbnail.png")

# sphinx_gallery_thumbnail_path = '_static/pcp_thumbnail.png'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a functional test for orion benchmark."""

import plotly
import pytest

from orion.benchmark.assessment import AverageRank, AverageResult
from orion.benchmark.benchmark_client import get_or_create_benchmark
from orion.benchmark.task import BaseTask, Branin, CarromTable, EggHolder, RosenBrock

algorithms = [
    {"algorithm": {"random": {"seed": 1}}},
    {"algorithm": {"tpe": {"seed": 1}}},
]


class BirdLike(BaseTask):
    """User defined benchmark task"""

    def __init__(self, max_trials=20):
        super(BirdLike, self).__init__(max_trials=max_trials)

    def call(self, x):

        y = (2 * x ** 4 + x ** 2 + 2) / (x ** 4 + 1)

        return [dict(name="birdlike", type="objective", value=y)]

    def get_search_space(self):
        """Return the search space for the task objective function"""
        rspace = {"x": "uniform(-4, 4)"}

        return rspace


@pytest.mark.usefixtures("setup_pickleddb_database")
def test_simple():
    """Test a end 2 end exucution of benchmark"""
    task_num = 2
    trial_num = 10
    assessments = [AverageResult(task_num), AverageRank(task_num)]
    tasks = [
        Branin(trial_num),
        BirdLike(trial_num),
    ]

    benchmark = get_or_create_benchmark(
        name="bm001",
        algorithms=algorithms,
        targets=[{"assess": assessments, "task": tasks}],
    )

    benchmark.process()

    assert len(benchmark.studies) == len(assessments) * len(tasks)

    status = benchmark.status()

    experiments = benchmark.experiments()

    assert len(experiments) == len(algorithms) * task_num * len(assessments) * len(
        tasks
    )

    assert len(status) == len(algorithms) * len(assessments) * len(tasks)

    figures = benchmark.analysis()

    assert len(figures) == len(benchmark.studies)
    assert type(figures[0]) is plotly.graph_objects.Figure

    benchmark = get_or_create_benchmark(name="bm001")
    figures = benchmark.analysis()

    assert len(figures) == len(benchmark.studies)
    assert type(figures[0]) is plotly.graph_objects.Figure



from orion.executor.base import BaseExecutor

try:
    from dask.distributed import Client, get_client, get_worker, rejoin, secede

    HAS_DASK = True
except ImportError:
    HAS_DASK = False


class Dask(BaseExecutor):
    def __init__(self, n_workers=-1, client=None, **config):
        super(Dask, self).__init__(n_workers=n_workers)

        if not HAS_DASK:
            raise ImportError("Dask must be installed to use Dask executor.")

        self.config = config
        if client is None:
            client = Client(**self.config)

        self.client = client

    def __getstate__(self):
        return super(Dask, self).__getstate__()

    def __setstate__(self, state):
        super(Dask, self).__setstate__(state)
        self.client = get_client()

    @property
    def in_worker(self):
        try:
            get_worker()
            return True
        except ValueError:
            return False

    def wait(self, futures):
        if self.in_worker:
            secede()
        results = self.client.gather(list(futures))
        if self.in_worker:
            rejoin()
        return results

    def submit(self, function, *args, **kwargs):
        return self.client.submit(function, *args, **kwargs, pure=False)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.client.close()
        super(Dask, self).__exit__(exc_type, exc_value, traceback)



# -*- coding: utf-8 -*-
"""Tests :func:`orion.analysis.partial_dependency`"""
import copy

import numpy
import pandas as pd
import pytest
from sklearn.ensemble import (
    AdaBoostRegressor,
    BaggingRegressor,
    ExtraTreesRegressor,
    GradientBoostingRegressor,
    RandomForestRegressor,
)

from orion.analysis.base import flatten_numpy, to_numpy, train_regressor
from orion.analysis.partial_dependency_utils import (
    make_grid,
    partial_dependency,
    partial_dependency_grid,
    reverse,
)
from orion.core.io.space_builder import SpaceBuilder
from orion.core.worker.transformer import build_required_space

data = pd.DataFrame(
    data={
        "id": ["a", "b", "c", "d"],
        "x": [0, 1, 2, 3],
        "y": list(numpy.e ** numpy.array([2, 3, 1, 4])),
        "z": ["a", "a", "b", "c"],
        "objective": [0.1, 0.2, 0.3, 0.5],
    }
)


hdata = pd.DataFrame(
    data={
        "id": ["a", "b", "c", "d"],
        "x": [0, 1, 2, 3],
        "y": list(numpy.e ** numpy.array([[2, 1, 2], [3, 0, 1], [1, 3, 0], [4, 4, 0]])),
        "z": ["a", "a", "b", "c"],
        "objective": [0.1, 0.2, 0.3, 0.5],
    }
)


@pytest.fixture
def space():
    """Base flat space"""
    return SpaceBuilder().build(
        {
            "x": "uniform(0, 6)",
            "y": "loguniform(0.001, 1)",
            "z": "choices(['a', 'b', 'c'])",
        }
    )


@pytest.fixture
def hspace():
    """Space with multidim dims"""
    return SpaceBuilder().build(
        {
            "x": "uniform(0, 6)",
            "y": f"loguniform({numpy.e}, {numpy.e}**4, shape=3)",
            "z": "choices(['a', 'b', 'c'])",
        }
    )


def flatten_space(some_space):
    """Flatten a space"""
    return build_required_space(
        some_space,
        dist_requirement="linear",
        type_requirement="numerical",
        shape_requirement="flattened",
    )


def mock_model():
    """Return a mocked regressor which just predict iterated integers"""

    class Model:
        """Mocked Regressor"""

        def __init__(self):
            self.i = 0

        def predict(self, data):
            """Returns counting of predictions requested."""
            data = numpy.arange(data.shape[0]) + self.i
            self.i += data.shape[0]
            return data  #  + numpy.random.normal(0, self.i, size=data.shape[0])

    return Model()


def mock_train_regressor(monkeypatch, assert_model=None):
    """Mock the train_regressor to return the mocked regressor instead"""

    def train_regressor(model, data, **kwargs):
        """Return the mocked model, and then model argument if requested"""
        if assert_model:
            assert model == assert_model
        return mock_model()

    monkeypatch.setattr(
        "orion.analysis.partial_dependency_utils.train_regressor", train_regressor
    )


def test_reverse_no_reshape_needed(space):
    """Test that transformed grids can be reversed to original type, when there was no reshape"""
    flattened_space = flatten_space(space)

    grid = {
        "x": [0, 1, 2, 4],
        "y": numpy.log([0.001, 0.01, 0.1, 1]),
        "z": [0, 0, 1, 2],
    }

    reversed_grid = reverse(flattened_space, grid)
    assert reversed_grid["x"] == [0, 1, 2, 4]
    assert reversed_grid["y"] == [0.001, 0.01, 0.1, 1]
    assert reversed_grid["z"] == ["a", "a", "b", "c"]


def test_reverse_with_reshape(hspace):
    """Test that transformed grids can be reversed to original type, when reshape was applied"""
    flattened_space = flatten_space(hspace)

    grid = {
        "x": [0, 1, 2, 4],
        "y[0]": numpy.log([0.001, 0.01, 0.1, 1]),
        "y[1]": numpy.log([1, 0.01, 0.1, 0.001]),
        "z": [0, 0, 1, 2],
    }

    reversed_grid = reverse(flattened_space, grid)
    assert reversed_grid["x"] == [0, 1, 2, 4]
    assert reversed_grid["y[0]"] == [0.001, 0.01, 0.1, 1]
    assert reversed_grid["y[1]"] == [1, 0.01, 0.1, 0.001]
    assert reversed_grid["z"] == ["a", "a", "b", "c"]


def test_make_grid(hspace):
    """Test that different dimension types are supported"""
    flattened_space = flatten_space(hspace)
    assert list(make_grid(flattened_space["x"], 7)) == [0, 1, 2, 3, 4, 5, 6]
    numpy.testing.assert_almost_equal(
        make_grid(flattened_space["y[0]"], 4), [1, 2, 3, 4], decimal=4
    )
    numpy.testing.assert_almost_equal(
        make_grid(flattened_space["y[2]"], 4), [1, 2, 3, 4], decimal=4
    )
    assert list(make_grid(flattened_space["z"], 10)) == [0, 1, 2]


def test_partial_dependency_grid(hspace):
    """Test the computation of the averages and stds"""

    flattened_space = flatten_space(hspace)

    n_points = 5
    n_samples = 20
    samples = flattened_space.sample(n_samples)
    samples = pd.DataFrame(samples, columns=flattened_space.keys())

    params = ["x", "y[0]", "y[2]", "z"]

    # Test for 1 param
    grid, averages, stds = partial_dependency_grid(
        flattened_space, mock_model(), ["x"], samples, n_points=n_points
    )

    assert list(grid.keys()) == ["x"]
    assert list(grid["x"]) == [0, 1.5, 3, 4.5, 6]
    assert averages.shape == (n_points,)
    assert stds.shape == (n_points,)
    assert averages[0] == numpy.arange(n_samples).mean()
    assert (
        averages[4]
        == numpy.arange(n_samples * (n_points - 1), n_samples * n_points).mean()
    )
    assert stds[0] == numpy.arange(n_samples).std()

    # Test for 2 param
    grid, averages, stds = partial_dependency_grid(
        flattened_space, mock_model(), ["x", "y[0]"], samples, n_points=n_points
    )

    assert list(grid.keys()) == ["x", "y[0]"]
    assert list(grid["x"]) == [0, 1.5, 3, 4.5, 6]
    # assert list(grid["y[0]"]) == [0, 0.75, 1.5, 2.25, 3]
    numpy.testing.assert_almost_equal(
        grid["y[0]"],
        numpy.linspace(
            numpy.log(float(f"{numpy.e}")),
            numpy.log(float(f"{numpy.e}") ** 4),
            num=n_points,
        ),
        decimal=4,
    )

    assert averages.shape == (n_points, n_points)
    assert stds.shape == (n_points, n_points)
    assert averages[0, 0] == numpy.arange(n_samples).mean()
    assert (
        averages[4, 4]
        == numpy.arange(
            n_samples * n_points * n_points - n_samples, n_samples * n_points * n_points
        ).mean()
    )
    assert stds[0, 0] == numpy.arange(n_samples).std()
    assert stds[4, 4] == numpy.arange(n_samples).std()

    # Test for 2 param with one categorical, with less categories then n_points
    grid, averages, stds = partial_dependency_grid(
        flattened_space, mock_model(), ["x", "z"], samples, n_points=n_points
    )

    assert list(grid.keys()) == ["x", "z"]
    assert list(grid["x"]) == [0, 1.5, 3, 4.5, 6]
    assert list(grid["z"]) == [0, 1, 2]
    assert averages.shape == (3, n_points)
    assert stds.shape == (3, n_points)

    assert averages[0, 0] == numpy.arange(n_samples).mean()
    assert (
        averages[2, 4]
        == numpy.arange(
            n_samples * 3 * n_points - n_samples, n_samples * 3 * n_points
        ).mean()
    )
    assert stds[0, 0] == numpy.arange(n_samples).std()
    assert stds[2, 4] == numpy.arange(n_samples).std()


def test_accept_empty(space):
    """Tests an empty dataframe is returned if you give an empty dataframe"""
    empty_frame = pd.DataFrame()
    results = partial_dependency(empty_frame, space, n_grid_points=3, n_samples=5)

    assert results == {}

    empty_frame = pd.DataFrame(columns=["x", "y", "z", "objective"])
    results = partial_dependency(empty_frame, space, n_grid_points=3, n_samples=5)

    assert results == {}


def test_parameter_not_modified(monkeypatch, space):
    """Tests the original dataframe is not modified"""
    original = copy.deepcopy(data)

    mock_train_regressor(monkeypatch)
    partial_dependency(data, space, n_samples=10)

    pd.testing.assert_frame_equal(data, original)


def test_single_param(monkeypatch, space):
    """Test computing for a single param"""
    mock_train_regressor(monkeypatch)

    partial_dependencies = partial_dependency(data, space, params=["x"], n_samples=10)
    assert list(partial_dependencies.keys()) == ["x"]
    assert len(partial_dependencies["x"][0]["x"]) == 10
    assert partial_dependencies["x"][1].shape == (10,)
    assert partial_dependencies["x"][2].shape == (10,)


def test_multiple_params(monkeypatch, space):
    """Test computing for multiple param"""
    mock_train_regressor(monkeypatch)

    partial_dependencies = partial_dependency(
        data, space, params=["x", "y"], n_samples=10
    )
    assert set(partial_dependencies.keys()) == {"x", ("x", "y"), "y"}
    assert len(partial_dependencies["x"][0]["x"]) == 10
    assert len(partial_dependencies["y"][0]["y"]) == 10
    assert len(partial_dependencies[("x", "y")][0]["x"]) == 10
    assert len(partial_dependencies[("x", "y")][0]["y"]) == 10
    assert partial_dependencies["x"][1].shape == (10,)
    assert partial_dependencies["x"][2].shape == (10,)
    assert partial_dependencies["y"][1].shape == (10,)
    assert partial_dependencies["y"][2].shape == (10,)
    assert partial_dependencies[("x", "y")][1].shape == (10, 10)
    assert partial_dependencies[("x", "y")][2].shape == (10, 10)


def test_multidim(monkeypatch, hspace):
    """Test computing for multiple params in multidim space"""
    mock_train_regressor(monkeypatch)

    partial_dependencies = partial_dependency(
        hdata, hspace, params=["x", "y[0]", "y[1]"], n_samples=10
    )
    assert set(partial_dependencies.keys()) == {
        "x",
        "y[0]",
        "y[1]",
        ("x", "y[0]"),
        ("x", "y[1]"),
        ("y[0]", "y[1]"),
    }
    assert len(partial_dependencies["x"][0]["x"]) == 10
    assert len(partial_dependencies["y[0]"][0]["y[0]"]) == 10
    assert len(partial_dependencies["y[1]"][0]["y[1]"]) == 10
    assert len(partial_dependencies[("x", "y[0]")][0]["x"]) == 10
    assert len(partial_dependencies[("x", "y[0]")][0]["y[0]"]) == 10
    assert partial_dependencies["x"][1].shape == (10,)
    assert partial_dependencies["x"][2].shape == (10,)
    assert partial_dependencies["y[0]"][1].shape == (10,)
    assert partial_dependencies["y[0]"][2].shape == (10,)
    assert partial_dependencies[("x", "y[0]")][1].shape == (10, 10)
    assert partial_dependencies[("x", "y[0]")][2].shape == (10, 10)


def test_categorical(monkeypatch, space):
    """Test computing for categorical"""
    mock_train_regressor(monkeypatch)

    partial_dependencies = partial_dependency(data, space, params=["z"], n_samples=10)

    assert set(partial_dependencies.keys()) == {"z"}
    assert partial_dependencies["z"][0]["z"] == ["a", "b", "c"]
    assert partial_dependencies["z"][1].shape == (3,)
    assert partial_dependencies["z"][2].shape == (3,)


def test_n_grid_points(monkeypatch, space):
    """Test that number of points on grid is correct, including adjustment for categorical dims"""
    mock_train_regressor(monkeypatch)

    n_grid_points = 5
    partial_dependencies = partial_dependency(
        data, space, params=["x", "z"], n_grid_points=n_grid_points, n_samples=10
    )

    assert set(partial_dependencies.keys()) == {"x", ("x", "z"), "z"}
    assert partial_dependencies["z"][1].shape == (3,)
    assert partial_dependencies["z"][2].shape == (3,)

    assert len(partial_dependencies["x"][0]["x"]) == n_grid_points
    assert len(partial_dependencies[("x", "z")][0]["x"]) == n_grid_points
    assert partial_dependencies[("x", "z")][0]["z"] == ["a", "b", "c"]
    assert partial_dependencies["z"][0]["z"] == ["a", "b", "c"]
    assert partial_dependencies["x"][1].shape == (n_grid_points,)
    assert partial_dependencies["x"][2].shape == (n_grid_points,)
    assert partial_dependencies["z"][1].shape == (3,)
    assert partial_dependencies["z"][2].shape == (3,)
    assert partial_dependencies[("x", "z")][1].shape == (3, n_grid_points)
    assert partial_dependencies[("x", "z")][2].shape == (3, n_grid_points)


def test_n_samples(monkeypatch, space):
    """Test n_samples is used properly when sampling random points"""
    mock_train_regressor(monkeypatch)

    PARAMS = ["x", "z"]
    N_SAMPLES = numpy.random.randint(20, 50)

    def mock_partial_dependency_grid(space, model, params, samples, n_points):
        assert samples.shape == (N_SAMPLES, len(space))
        return partial_dependency_grid(space, model, params, samples, n_points)

    monkeypatch.setattr(
        "orion.analysis.partial_dependency_utils.partial_dependency_grid",
        mock_partial_dependency_grid,
    )

    partial_dependency(data, space, params=PARAMS, n_grid_points=5, n_samples=N_SAMPLES)



# -*- coding: utf-8 -*-
"""
Provide tools to calculate Local Parameter Importance
=====================================================
"""
import numpy
import pandas as pd

from orion.analysis.base import flatten_numpy, to_numpy, train_regressor
from orion.core.worker.transformer import build_required_space


def make_grid(point, space, model, n_points):
    """Build a grid based on point.

    The shape of the grid will be
        (number of hyperparameters,
         number of points ``n_points``,
         number of hyperparameters + 1)

    Last column is the objective predicted by the model for a given point.

    Parameters
    ----------
    point: numpy.ndarray
        A tuple representation of the best trials, (hyperparameters + objective)
    space: Space object
        A space object from an experiment. It must be flattened and linearized.
    model: `sklearn.base.RegressorMixin`
        Trained regressor used to compute predictions on the grid
    n_points: int
        Number of points for each dimension on the grid.

    """
    grid = numpy.zeros((len(space), n_points, len(space) + 1))
    for i, dim in enumerate(space.values()):
        grid[i, :, :] = point
        grid[i, :, i] = numpy.linspace(*dim.interval(), num=n_points)
        grid[i, :, -1] = model.predict(grid[i, :, :-1])
    return grid


def compute_variances(grid):
    """Compute variance for each hyperparameters"""
    return grid[:, :, -1].var(axis=1)


def _lpi(point, space, model, n_points):
    """Local parameter importance for each hyperparameters"""
    grid = make_grid(point, space, model, n_points)
    variances = compute_variances(grid)
    ratios = variances / variances.sum()
    return ratios


# def _linear_lpi(point, space, model, n):
#     # TODO
#     return


modes = dict(best=_lpi)  # , linear=_linear_lpi)


def lpi(
    trials,
    space,
    mode="best",
    model="RandomForestRegressor",
    n_points=20,
    n_runs=10,
    **kwargs
):
    """
    Calculates the Local Parameter Importance for a collection of
    :class:`orion.core.worker.trial.Trial`.

    For more information on the metric, see original paper at
    https://ml.informatik.uni-freiburg.de/papers/18-LION12-CAVE.pdf.

    Biedenkapp, André, et al. "Cave: Configuration assessment, visualization and evaluation."
    International Conference on Learning and Intelligent Optimization. Springer, Cham, 2018.

    Parameters
    ----------
    trials: DataFrame or dict
        A dataframe of trials containing, at least, the columns 'objective' and 'id'. Or a dict
        equivalent.

    space: Space object
        A space object from an experiment.

    mode: str
        Mode to compute the LPI.
        - ``best``: Take the best trial found as the anchor for the LPI
        - ``linear``: Recompute LPI for all values on a grid

    model: str
        Name of the regression model to use. Can be one of
        - AdaBoostRegressor
        - BaggingRegressor
        - ExtraTreesRegressor
        - GradientBoostingRegressor
        - RandomForestRegressor (Default)

    n_points: int
        Number of points to compute the variances. Default is 20.

    n_runs: int
        Number of runs to compute the standard error of the LPI. Default is 10.

    ``**kwargs``
        Arguments for the regressor model.

    Returns
    -------
    DataFrame
        LPI value for each parameter. If ``mode`` is `linear`, then a list of
        param values and LPI metrics are returned in a DataFrame format.

    """
    flattened_space = build_required_space(
        space,
        dist_requirement="linear",
        type_requirement="numerical",
        shape_requirement="flattened",
    )
    if trials.empty or trials.shape[0] == 0:
        return pd.DataFrame(
            data=[0] * len(flattened_space),
            index=flattened_space.keys(),
            columns=["LPI"],
        )

    data = to_numpy(trials, space)
    data = flatten_numpy(data, flattened_space)
    best_point = data[numpy.argmin(data[:, -1])]
    rng = numpy.random.RandomState(kwargs.pop("random_state", None))
    results = numpy.zeros((n_runs, len(flattened_space)))
    for i in range(n_runs):
        trained_model = train_regressor(
            model, data, random_state=rng.randint(2 ** 32 - 1), **kwargs
        )
        results[i] = modes[mode](best_point, flattened_space, trained_model, n_points)

    averages = results.mean(0)
    standard_errors = results.std(0)
    frame = pd.DataFrame(
        data=numpy.array([averages, standard_errors]).T,
        index=flattened_space.keys(),
        columns=["LPI", "STD"],
    )
    return frame



"""
Base tools to compute diverse analysis
======================================

"""
import itertools

import numpy
from sklearn.ensemble import (
    AdaBoostRegressor,
    BaggingRegressor,
    ExtraTreesRegressor,
    GradientBoostingRegressor,
    RandomForestRegressor,
)

from orion.core.worker.transformer import build_required_space

_regressors_ = {
    "AdaBoostRegressor": AdaBoostRegressor,
    "BaggingRegressor": BaggingRegressor,
    "ExtraTreesRegressor": ExtraTreesRegressor,
    "GradientBoostingRegressor": GradientBoostingRegressor,
    "RandomForestRegressor": RandomForestRegressor,
}


def average(trials, group_by="order", key="best", return_var=False):
    """Compute the average of some trial attribute.

    By default it will compute the average objective at each time step across
    multiple experiments.

    Parameters
    ----------
    trials: DataFrame
        A dataframe of trials containing, at least, the columns 'best' and 'order'.
    group_by: str, optional
        The attribute to use to group trials for the average. By default it group trials
        by order (ex: all first trials across experiments.)
    key: str, optional
        The attribute to average. Defaults to 'best' as returned by ``orion.analysis.regret``.
    return_var: bool, optional
        If True, and a column '{key}_var' where '{key}' is the value of the argument `key`.
        Defaults to False.

    Returns
    -------
    A dataframe with columns 'order', '{key}_mean' and '{key}_var'.

    """
    if trials.empty:
        return trials

    group = trials.groupby(group_by)
    mean = group[key].mean().reset_index().rename(columns={key: f"{key}_mean"})
    if return_var:
        mean[f"{key}_var"] = group[key].var().reset_index()[key]

    return mean


# pylint:disable=unsupported-assignment-operation
def ranking(trials, group_by="order", key="best"):
    """Compute the ranking of some trial attribute.

    By default it will compute the ranking with respect to objectives at each time step across
    multiple experiments.

    Parameters
    ----------
    trials: DataFrame
        A dataframe of trials containing, at least, the columns 'best' and 'order'.
    group_by: str, optional
        The attribute to use to group trials for the ranking. By default it group trials
        by order (ex: all first trials across experiments.)
    key: str, optional
        The attribute to use for the ranking. Defaults to 'best' as returned by
        ``orion.analysis.regret``.

    Returns
    -------
    A copy of the original dataframe with a new column 'rank' for the rankings.

    """
    if trials.empty:
        return trials

    def rank(row):
        indices = row[key].argsort().to_numpy()
        ranks = numpy.empty_like(indices)
        ranks[indices] = numpy.arange(len(ranks))
        row["rank"] = ranks
        return row

    return trials.groupby(group_by).apply(rank)


def flatten_params(space, params=None):
    """Return the params of the corresponding flat space

    If no params are passed, returns all flattened params.
    If params are passed, returns the corresponding flattened params.

    Parameters
    ----------
    space: Space object
        A space object from an experiment.
    params: list of str, optional
        The parameters to select from the search space. If the flattened search space
        contains flattened params such as ('y' -> 'y[0]', 'y[1]'), passing 'y' in the list of
        params will returned the flattened version ['y[0]', 'y[1]']

    Examples
    --------
    If space has x~uniform(0, 1) and y~uniform(0, 1, shape=(1, 2)).
    >>> flatten_params(space)
    ['x', 'y[0,0]', 'y[0,1]']
    >>> flatten_params(space, params=['x'])
    ['x']
    >>> flatten_params(space, params=['x', 'y'])
    ['x', 'y[0,0]', 'y[0,1]']
    >>> flatten_params(space, params=['x', 'y[0,1]'])
    ['x', 'y[0,1]']
    >>> flatten_params(space, params=['y[0,1]', 'x'])
    ['x', 'y[0,1]']

    Raises
    ------
    ValueError
        If one of the parameter names passed is not in the flattened space.

    """
    keys = set(space.keys())
    flattened_keys = set(
        build_required_space(
            space,
            dist_requirement="linear",
            type_requirement="numerical",
            shape_requirement="flattened",
        ).keys()
    )

    if params is None:
        return sorted(flattened_keys)

    flattened_params = []
    for param in params:
        if param not in flattened_keys and param not in keys:
            raise ValueError(
                f"Parameter {param} not contained in space: {flattened_keys}"
            )
        elif param not in flattened_keys and param in keys:
            dim = space[param]
            flattened_params += [
                f'{dim.name}[{",".join(map(str, index))}]'
                for index in itertools.product(*map(range, dim.shape))
            ]
        else:
            flattened_params.append(param)

    return flattened_params


def to_numpy(trials, space):
    """Convert trials in DataFrame to Numpy array of (params + objective)"""
    return trials[list(space.keys()) + ["objective"]].to_numpy()


def flatten_numpy(trials_array, flattened_space):
    """Flatten dimensions"""
    flattened_points = numpy.array(
        [flattened_space.transform(point[:-1]) for point in trials_array]
    )

    return numpy.concatenate((flattened_points, trials_array[:, -1:]), axis=1)


def train_regressor(regressor_name, data, **kwargs):
    """Train regressor model

    Parameters
    ----------
    model: str
        Name of the regression model to use. Can be one of
        - AdaBoostRegressor
        - BaggingRegressor
        - ExtraTreesRegressor
        - GradientBoostingRegressor
        - RandomForestRegressor (Default)
    trials: DataFrame or dict
        A dataframe of trials containing, at least, the columns 'objective' and 'id'. Or a dict
        equivalent.

    **kwargs
        Arguments for the regressor model.

    """
    if regressor_name not in _regressors_:
        raise ValueError(
            f"{regressor_name} is not a supported regressor. "
            f"Did you mean any of theses: list(_regressors_.keys())"
        )

    regressor = _regressors_[regressor_name](**kwargs)
    return regressor.fit(data[:, :-1], data[:, -1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Task for CarromTable Function
==============================
"""
import numpy

from orion.benchmark.task.base import BaseTask


class CarromTable(BaseTask):
    """`CarromTable function <http://infinity77.net/global_optimization/test_functions_nd_C.html#go_benchmark.CarromTable>`_
    as benchmark task"""

    def __init__(self, max_trials=20):
        super(CarromTable, self).__init__(max_trials=max_trials)

    def call(self, x):
        """Evaluate a 2-D CarromTable function."""
        a = numpy.exp(
            2 * numpy.absolute(1 - numpy.sqrt(numpy.square(x[0]) + numpy.square(x[1])))
        )
        y = -a / 30 * numpy.square(numpy.cos(x[0])) * numpy.square(numpy.cos(x[1]))

        return [dict(name="carromtable", type="objective", value=y)]

    def get_search_space(self):
        """Return the search space for the task objective function"""
        rspace = {"x": "uniform(-10, 10, shape=2)"}

        return rspace



"""Tests for :mod:`orion.analysis.base`"""
import numpy
import pandas as pd
import pytest
from sklearn.ensemble import (
    AdaBoostRegressor,
    BaggingRegressor,
    ExtraTreesRegressor,
    GradientBoostingRegressor,
    RandomForestRegressor,
)

from orion.analysis.base import (
    average,
    flatten_numpy,
    flatten_params,
    ranking,
    to_numpy,
    train_regressor,
)
from orion.core.io.space_builder import SpaceBuilder
from orion.core.worker.transformer import build_required_space

data = pd.DataFrame(
    data={
        "id": ["a", "b", "c"],
        "x": [0, 1, 2],
        "y": [1, 2, 0],
        "z": ["a", "b", "c"],
        "objective": [0.1, 0.2, 0.3],
    }
)


@pytest.fixture
def space():
    space = SpaceBuilder().build(
        {
            "x": "uniform(0, 6)",
            "y": "uniform(0, 3)",
            "z": "choices(['a', 'b', 'c'])",
        }
    )
    return space


@pytest.fixture
def fspace(space):
    return build_required_space(
        space,
        dist_requirement="linear",
        type_requirement="numerical",
        shape_requirement="flattened",
    )


@pytest.fixture
def params():
    return list("xyz")


@pytest.fixture
def hspace():
    space = SpaceBuilder().build(
        {
            "x": "uniform(0, 6, shape=3)",
            "y": "uniform(0, 3, shape=(2, 3))",
            "z": "choices(['a', 'b', 'c'])",
        }
    )
    return space


@pytest.fixture
def flat_params():
    return (
        [f"x[{i}]" for i in range(3)]
        + [f"y[{i},{j}]" for i in range(2) for j in range(3)]
        + ["z"]
    )


class TestAverage:
    def test_accept_empty(self):
        """Tests an empty dataframe is returned if you give an empty dataframe"""
        empty_frame = pd.DataFrame()
        result = average(empty_frame)

        assert result.empty
        assert result.equals(empty_frame)

        empty_frame = pd.DataFrame(columns=["order", "best"])
        result = average(empty_frame)

        assert result.empty
        assert "best_mean" not in result.columns
        assert "best_var" not in result.columns
        assert result.equals(empty_frame)

    def test_parameter_not_modified(self):
        """Tests the original dataframe is not modified"""
        data = pd.DataFrame(
            data={"order": ["a", "b", "c", "d"], "best": [0.1, 0.2, 0.3, 0.5]}
        )

        result = average(data)

        assert data.columns.tolist() == ["order", "best"]
        assert result.columns.tolist() == ["order", "best_mean"]

    def test_average_statistic(self):
        """Test that the average is correctly computed"""
        data = pd.DataFrame(
            data={
                "id": [0, 0, 0, 0, 1, 1, 1, 1],
                "order": [0, 1, 2, 3, 0, 1, 2, 3],
                "best": [0.1, 0.2, 0.3, 0.5, 0.3, 0.2, 0.5, 0.6],
            }
        )

        result = average(data)

        ref = (data["best"][:4].to_numpy() + data["best"][4:].to_numpy()) / 2
        numpy.testing.assert_equal(result["best_mean"].to_numpy(), ref)

    def test_variance_statistic(self):
        """Test that the variance is correctly computed"""
        data = pd.DataFrame(
            data={
                "id": [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2],
                "order": [0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3],
                "best": [0.1, 0.2, 0.3, 0.5, 0.3, 0.2, 0.5, 0.6, 0.3, 0.4, 0.1, 0.3],
            }
        )

        result = average(data, return_var=True)

        ref = data["best"].to_numpy().reshape((3, 4)).var(axis=0, ddof=1)

        numpy.testing.assert_allclose(
            result["best_var"].to_numpy(),
            ref,
        )

    def test_unbalanced_groups(self):
        """Test that the mean is correctly computed on unbalanced groups"""
        data = pd.DataFrame(
            data={
                "id": [0, 0, 0, 0, 1, 1, 1, 2, 2],
                "order": [0, 1, 2, 3, 0, 1, 2, 0, 1],
                "best": [0.1, 0.2, 0.3, 0.5, 0.3, 0.2, 0.5, 0.3, 0.4],
            }
        )

        result = average(data, return_var=True)

        assert (
            result["best_mean"][0]
            == (data["best"][0] + data["best"][4] + data["best"][7]) / 3
        )

        assert (
            result["best_mean"][1]
            == (data["best"][1] + data["best"][5] + data["best"][8]) / 3
        )

        assert result["best_mean"][2] == (data["best"][2] + data["best"][6]) / 2

        assert result["best_mean"][3] == data["best"][3]

    def test_custom_group_by(self):
        """Test using custom group_by argument"""
        data = pd.DataFrame(
            data={
                "step": ["a", "b", "c", "d"],
                "best": [0.1, 0.2, 0.3, 0.5],
            }
        )

        result = average(data, group_by="step")

        assert result.columns.tolist() == ["step", "best_mean"]

    def test_custom_key(self):
        """Test using custom key argument"""
        data = pd.DataFrame(
            data={
                "order": ["a", "b", "c", "d"],
                "objective": [0.1, 0.2, 0.3, 0.5],
            }
        )

        result = average(data, key="objective")

        assert result.columns.tolist() == ["order", "objective_mean"]


class TestRanking:
    def test_accept_empty(self):
        """Tests an empty dataframe is returned if you give an empty dataframe"""
        empty_frame = pd.DataFrame()
        result = ranking(empty_frame)

        assert result.empty
        assert result.equals(empty_frame)

        empty_frame = pd.DataFrame(columns=["order", "best"])
        result = ranking(empty_frame)

        assert result.empty
        assert "rank" not in result.columns
        assert result.equals(empty_frame)

    def test_parameter_not_modified(self):
        """Tests the original dataframe is not modified"""
        data = pd.DataFrame(
            data={"order": ["a", "b", "c", "d"], "best": [0.1, 0.2, 0.3, 0.5]}
        )

        result = ranking(data)

        assert data.columns.tolist() == ["order", "best"]
        assert result.columns.tolist() == ["order", "best", "rank"]

    def test_ranking(self):
        """Test that the average is correctly computed"""
        data = pd.DataFrame(
            data={
                "id": [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2],
                "order": [0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3],
                "best": [0.1, 0.2, 0.3, 0.5, 0.3, 0.2, 0.5, 0.6, 0.3, 0.4, 0.1, 0.3],
            }
        )
        result = ranking(data)

        ref = data["best"].to_numpy().reshape((3, 4))

        assert result["rank"].tolist() == [0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 0, 0]

    def test_unbalanced_groups(self):
        """Test that the mean is correctly computed on unbalanced groups"""
        data = pd.DataFrame(
            data={
                "id": [0, 0, 0, 0, 1, 1, 1, 2, 2],
                "order": [0, 1, 2, 3, 0, 1, 2, 0, 1],
                "best": [0.1, 0.7, 0.6, 0.5, 0.4, 0.2, 0.5, 0.3, 0.1],
            }
        )

        result = ranking(data)

        assert result["rank"].tolist() == [0, 2, 1, 0, 2, 1, 0, 1, 0]

    def test_custom_group_by(self):
        """Test using custom group_by argument"""
        data = pd.DataFrame(
            data={
                "step": ["a", "b", "c", "d"],
                "best": [0.1, 0.2, 0.3, 0.5],
            }
        )

        result = ranking(data, group_by="step")

        assert result.columns.tolist() == ["step", "best", "rank"]

    def test_custom_key(self):
        """Test using custom key argument"""
        data = pd.DataFrame(
            data={
                "order": ["a", "b", "c", "d"],
                "objective": [0.1, 0.2, 0.3, 0.5],
            }
        )

        result = ranking(data, key="objective")

        assert result.columns.tolist() == ["order", "objective", "rank"]


class TestFlattenParams:
    def test_no_params(self, space, params):
        """Test that all params are returned if None"""
        assert flatten_params(space) == params

    def test_flat_no_params(self, hspace, flat_params):
        """Test that all flattened params are returned if None"""
        assert flatten_params(hspace) == flat_params

    def test_params_unchanged(self, space):
        """Test that params list passed is not modified"""
        params = ["x", "y"]
        flatten_params(space, params)
        assert params == ["x", "y"]

    def test_unexisting_params(self, space):
        """Test that ValueError is raised if passing unexisting params"""
        with pytest.raises(ValueError) as exc:
            flatten_params(space, ["idoexistbelieveme!!!"])
        assert exc.match(f"Parameter idoexistbelieveme!!! not contained in space: ")

    def test_no_flatten(self, space, hspace):
        """Test selection of params not involving flattening"""
        assert flatten_params(space, ["x", "y"]) == ["x", "y"]
        assert flatten_params(hspace, ["z"]) == ["z"]

    def test_flattened_params(self, hspace):
        """Test selecting specific flattened params"""
        params = ["x[0]", "x[2]", "y[0,2]", "y[1,1]", "z"]
        assert flatten_params(hspace, params) == params

    def test_top_params(self, hspace):
        """Test selecting all flattened keys of a parameter"""
        params = ["x", "y[0,2]", "y[1,1]", "z"]
        assert (
            flatten_params(hspace, params) == [f"x[{i}]" for i in range(3)] + params[1:]
        )


def test_to_numpy(space):
    """Test that trials are correctly converted to numpy array"""
    array = to_numpy(data, space)

    assert array.shape == (3, 4)
    numpy.testing.assert_equal(array[:, 0], data["x"])
    numpy.testing.assert_equal(array[:, 1], data["y"])
    numpy.testing.assert_equal(array[:, 2], data["z"])
    numpy.testing.assert_equal(array[:, 3], data["objective"])


class TestTrainRegressor:
    def test_train_regressor(self, space, fspace):
        """Test training different models"""
        array = flatten_numpy(to_numpy(data, space), fspace)
        model = train_regressor("AdaBoostRegressor", array)
        assert isinstance(model, AdaBoostRegressor)
        model = train_regressor("BaggingRegressor", array)
        assert isinstance(model, BaggingRegressor)
        model = train_regressor("ExtraTreesRegressor", array)
        assert isinstance(model, ExtraTreesRegressor)
        model = train_regressor("GradientBoostingRegressor", array)
        assert isinstance(model, GradientBoostingRegressor)
        model = train_regressor("RandomForestRegressor", array)
        assert isinstance(model, RandomForestRegressor)

    def test_train_regressor_kwargs(self, space, fspace):
        """Test training models with kwargs"""
        array = flatten_numpy(to_numpy(data, space), fspace)
        model = train_regressor(
            "RandomForestRegressor", array, max_depth=2, max_features="sqrt"
        )
        assert model.max_depth == 2
        assert model.max_features == "sqrt"

    def test_train_regressor_invalid(self, space, fspace):
        """Test error message for invalid model names"""
        array = flatten_numpy(to_numpy(data, space), fspace)
        with pytest.raises(ValueError) as exc:
            train_regressor("IDontExist", array)
        assert exc.match("IDontExist is not a supported regressor")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.client.experiment`."""
import copy
import datetime
import logging

import joblib
import pandas.testing
import pytest

import orion.core
from orion.core.io.database import DuplicateKeyError
from orion.core.utils.exceptions import (
    BrokenExperiment,
    CompletedExperiment,
    SampleTimeout,
)
from orion.core.worker.trial import Trial
from orion.executor.base import Executor
from orion.storage.base import get_storage
from orion.testing import create_experiment, mock_space_iterate

config = dict(
    name="supernaekei",
    space={"x": "uniform(0, 200)"},
    metadata={
        "user": "tsirif",
        "orion_version": "XYZ",
        "VCS": {
            "type": "git",
            "is_dirty": False,
            "HEAD_sha": "test",
            "active_branch": None,
            "diff_sha": "diff",
        },
    },
    version=1,
    pool_size=1,
    max_trials=10,
    max_broken=5,
    working_dir="",
    algorithms={"random": {"seed": 1}},
    producer={"strategy": "NoParallelStrategy"},
    refers=dict(root_id="supernaekei", parent_id=None, adapter=[]),
)


base_trial = {
    "experiment": 0,
    "status": "new",  # new, reserved, suspended, completed, broken
    "worker": None,
    "start_time": None,
    "end_time": None,
    "heartbeat": None,
    "results": [],
    "params": [],
}


def compare_trials(trials_a, trials_b):
    """Compare two trials by using their configuration"""

    def to_dict(trial):
        return trial.to_dict()

    assert list(map(to_dict, trials_a)) == list(map(to_dict, trials_b))


def compare_without_heartbeat(trial_a, trial_b):
    """Compare trials configuration ommiting heartbeat"""
    trial_a_dict = trial_a.to_dict()
    trial_b_dict = trial_b.to_dict()
    trial_a_dict.pop("heartbeat")
    trial_b_dict.pop("heartbeat")
    assert trial_a_dict == trial_b_dict


def test_plot_is_defined():
    """Tests plot() method is defined"""
    with create_experiment(config, base_trial) as (_, _, client):
        assert client.plot()


def test_experiment_fetch_trials():
    """Test compliance of client and experiment `fetch_trials()`"""
    with create_experiment(config, base_trial) as (cfg, experiment, client):
        assert len(experiment.fetch_trials()) == 5
        compare_trials(experiment.fetch_trials(), client.fetch_trials())


def test_experiment_get_trial():
    """Test compliance of client and experiment `get_trial()`"""
    with create_experiment(config, base_trial) as (cfg, experiment, client):
        assert experiment.get_trial(uid=0) == client.get_trial(uid=0)


def test_experiment_fetch_trials_by_status():
    """Test compliance of client and experiment `fetch_trials_by_status()`"""
    with create_experiment(config, base_trial) as (cfg, experiment, client):
        compare_trials(
            experiment.fetch_trials_by_status("completed"),
            client.fetch_trials_by_status("completed"),
        )


def test_experiment_fetch_non_completed_trials():
    """Test compliance of client and experiment `fetch_noncompleted_trials()`"""
    with create_experiment(config, base_trial) as (cfg, experiment, client):
        compare_trials(
            experiment.fetch_noncompleted_trials(), client.fetch_noncompleted_trials()
        )


def test_experiment_to_pandas():
    """Test compliance of client and experiment `to_pandas()`"""
    with create_experiment(config, base_trial) as (cfg, experiment, client):
        pandas.testing.assert_frame_equal(experiment.to_pandas(), client.to_pandas())


@pytest.mark.usefixtures("version_XYZ")
class TestInsert:
    """Tests for ExperimentClient.insert"""

    def test_insert_params_wo_results(self):
        """Test insertion without results without reservation"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.insert(dict(x=100))
            assert trial.status == "interrupted"
            assert trial.params["x"] == 100
            assert trial.id in set(trial.id for trial in experiment.fetch_trials())
            compare_without_heartbeat(trial, client.get_trial(uid=trial.id))

            assert client._pacemakers == {}

    def test_insert_params_with_results(self):
        """Test insertion with results without reservation"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            timestamp = datetime.datetime.utcnow()
            trial = client.insert(
                dict(x=100), [dict(name="objective", type="objective", value=101)]
            )
            assert trial.status == "completed"
            assert trial.params["x"] == 100
            assert trial.objective.value == 101
            assert trial.end_time >= timestamp
            assert trial.id in set(trial.id for trial in experiment.fetch_trials())
            compare_without_heartbeat(trial, client.get_trial(uid=trial.id))
            assert client.get_trial(uid=trial.id).objective.value == 101

            assert client._pacemakers == {}

    def test_insert_params_with_results_and_reserve(self):
        """Test insertion with results and reservation"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            with pytest.raises(ValueError) as exc:
                client.insert(
                    dict(x=100),
                    [dict(name="objective", type="objective", value=101)],
                    reserve=True,
                )

            assert "Cannot observe a trial and reserve it" in str(exc.value)

    def test_insert_existing_params(self, monkeypatch):
        """Test that duplicated trials cannot be saved in storage"""
        mock_space_iterate(monkeypatch)
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            with pytest.raises(DuplicateKeyError) as exc:
                client.insert(dict(x=1))

            assert (
                "A trial with params {'x': 1} already exist for experiment supernaekei-v1"
                == str(exc.value)
            )

            assert client._pacemakers == {}

    def test_insert_partial_params(self):
        """Test that trial with missing dimension that has a default value can be saved"""
        config_with_default = copy.deepcopy(config)
        config_with_default["space"]["y"] = "uniform(0, 10, default_value=5)"
        trial_with_default = copy.deepcopy(base_trial)
        trial_with_default["params"].append({"name": "y", "type": "real", "value": 1})
        with create_experiment(config_with_default, trial_with_default) as (
            _,
            experiment,
            client,
        ):
            trial = client.insert(dict(x=100))

            assert trial.status == "interrupted"
            assert trial.params["x"] == 100
            assert trial.params["y"] == 5
            assert trial.id in set(trial.id for trial in experiment.fetch_trials())
            compare_without_heartbeat(trial, client.get_trial(uid=trial.id))

            assert client._pacemakers == {}

    def test_insert_partial_params_missing(self):
        """Test that trial with missing dimension cannot be saved"""
        config_with_default = copy.deepcopy(config)
        config_with_default["space"]["y"] = "uniform(0, 10)"
        trial_with_default = copy.deepcopy(base_trial)
        trial_with_default["params"].append({"name": "y", "type": "real", "value": 1})
        with create_experiment(config_with_default, trial_with_default) as (
            _,
            _,
            client,
        ):
            with pytest.raises(ValueError) as exc:
                client.insert(dict(x=1))

            assert (
                "Dimension y not specified and does not have a default value."
                == str(exc.value)
            )

    def test_insert_params_and_reserve(self):
        """Test that new trial is reserved properly with `reserve=True`"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.insert(dict(x=100), reserve=True)
            assert trial.status == "reserved"
            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()

    def test_insert_params_fails_not_reserved(self, monkeypatch):
        """Test that failed insertion because of duplicated trials will not reserve the original
        trial
        """
        mock_space_iterate(monkeypatch)
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            with pytest.raises(DuplicateKeyError):
                client.insert(dict(x=1), reserve=True)

            assert client._pacemakers == {}

    def test_insert_bad_params(self):
        """Test that bad params cannot be registered in storage"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            with pytest.raises(ValueError) as exc:
                client.insert(dict(x="bad bad bad"))

            assert (
                "Dimension x value bad bad bad is outside of prior uniform(0, 200)"
                == str(exc.value)
            )
            assert client._pacemakers == {}

    def test_insert_params_bad_results(self):
        """Test that results with from format cannot be saved (trial is registered anyhow)"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            with pytest.raises(ValueError) as exc:
                client.insert(
                    dict(x=100), [dict(name="objective", type="bad bad bad", value=0)]
                )

            assert "Given type, bad bad bad, not one of: " in str(exc.value)
            assert client._pacemakers == {}


@pytest.mark.usefixtures("version_XYZ")
class TestReserve:
    """Tests for ExperimentClient.reserve"""

    def test_reserve(self):
        """Test reservation of registered trials"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = experiment.get_trial(uid=cfg.trials[1]["_id"])
            assert trial.status != "reserved"
            client.reserve(trial)
            assert trial.status == "reserved"
            assert experiment.get_trial(trial).status == "reserved"
            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()

    def test_reserve_dont_exist(self):
        """Verify that unregistered trials cannot be reserved."""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(experiment="idontexist", params=cfg.trials[0]["params"])
            with pytest.raises(ValueError) as exc:
                client.reserve(trial)

            assert "Trial {} does not exist in database.".format(trial.id) == str(
                exc.value
            )
            assert client._pacemakers == {}

    def test_reserve_reserved_locally(self, caplog):
        """Verify that a trial cannot be reserved twice locally (warning, no exception)"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = experiment.get_trial(uid=cfg.trials[1]["_id"])
            assert trial.status != "reserved"
            client.reserve(trial)
            with caplog.at_level(logging.WARNING):
                client.reserve(trial)

            assert (
                "Trial {} is already reserved.".format(trial.id)
                == caplog.records[-1].message
            )

            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()

    def test_reserve_reserved_remotely(self):
        """Verify that a trial cannot be reserved if already reserved by another process"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(**cfg.trials[1])
            assert trial.status == "interrupted"
            client.reserve(trial)
            remote_pacemaker = client._pacemakers.pop(trial.id)
            assert experiment.get_trial(trial).status == "reserved"

            trial = Trial(**cfg.trials[1])
            assert trial.status == "interrupted"
            with pytest.raises(RuntimeError) as exc:
                client.reserve(trial)

            assert "Could not reserve trial {}.".format(trial.id) == str(exc.value)

            assert trial.status == "interrupted"
            assert experiment.get_trial(trial).status == "reserved"
            assert client._pacemakers == {}
            remote_pacemaker.stop()

    def test_reserve_race_condition(self):
        """Verify that race conditions during `reserve` is detected and raises a comprehensible
        error
        """
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.get_trial(uid=cfg.trials[0]["_id"])
            experiment.set_trial_status(trial, "reserved")
            trial.status = "new"  # Let's pretend it is still available

            with pytest.raises(RuntimeError) as exc:
                client.reserve(trial)

            assert "Could not reserve trial {}.".format(trial.id) == str(exc.value)
            assert client._pacemakers == {}


@pytest.mark.usefixtures("version_XYZ")
class TestRelease:
    """Tests for ExperimentClient.release"""

    def test_release(self):
        """Test releasing (to interrupted)"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = experiment.get_trial(uid=cfg.trials[1]["_id"])
            client.reserve(trial)
            pacemaker = client._pacemakers[trial.id]
            client.release(trial)
            assert trial.status == "interrupted"
            assert experiment.get_trial(trial).status == "interrupted"
            assert trial.id not in client._pacemakers
            assert not pacemaker.is_alive()

    def test_release_status(self):
        """Test releasing with a specific status"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = experiment.get_trial(uid=cfg.trials[1]["_id"])
            client.reserve(trial)
            pacemaker = client._pacemakers[trial.id]
            client.release(trial, "broken")
            assert trial.status == "broken"
            assert experiment.get_trial(trial).status == "broken"
            assert trial.id not in client._pacemakers
            assert not pacemaker.is_alive()

    def test_release_invalid_status(self):
        """Test releasing with a specific status"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = experiment.get_trial(uid=cfg.trials[1]["_id"])
            client.reserve(trial)
            with pytest.raises(ValueError) as exc:
                client.release(trial, "mouf mouf")

            assert exc.match("Given status `mouf mouf` not one of")

    def test_release_dont_exist(self, monkeypatch):
        """Verify that unregistered trials cannot be released"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(experiment="idontexist", params=cfg.trials[1]["params"])

            def do_nada(trial, **kwargs):
                """Don't do anything"""
                return None

            monkeypatch.setattr(client, "_release_reservation", do_nada)

            with pytest.raises(ValueError) as exc:
                client.release(trial)

            assert "Trial {} does not exist in database.".format(trial.id) == str(
                exc.value
            )
            assert client._pacemakers == {}

    def test_release_race_condition(self):
        """Verify that race conditions during `release` is detected and raises a comprehensible
        error
        """
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.get_trial(uid=cfg.trials[1]["_id"])
            client.reserve(trial)
            pacemaker = client._pacemakers[trial.id]
            # Woops! Trial got failed over from another process.
            experiment.set_trial_status(trial, "interrupted")
            trial.status = "reserved"  # Let's pretend we don't know.

            with pytest.raises(RuntimeError) as exc:
                client.release(trial)

            assert "Reservation for trial {} has been lost before release.".format(
                trial.id
            ) in str(exc.value)
            assert client._pacemakers == {}
            assert not pacemaker.is_alive()

    def test_release_unreserved(self):
        """Verify that unreserved trials cannot be released"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.get_trial(uid=cfg.trials[1]["_id"])
            with pytest.raises(RuntimeError) as exc:
                client.release(trial)

            assert "Trial {} was already released locally.".format(trial.id) == str(
                exc.value
            )

            assert client._pacemakers == {}

    def test_release_already_released_but_incorrectly(self):
        """Verify that incorrectly released trials have its pacemaker stopped properly"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.get_trial(uid=cfg.trials[1]["_id"])
            client.reserve(trial)
            pacemaker = client._pacemakers[trial.id]
            assert trial.status == "reserved"
            experiment.set_trial_status(trial, "interrupted")
            assert trial.status == "interrupted"

            with pytest.raises(RuntimeError) as exc:
                client.release(trial)

            assert "Trial {} was already released locally.".format(trial.id) == str(
                exc.value
            )

            assert client._pacemakers == {}
            assert not pacemaker.is_alive()


@pytest.mark.usefixtures("version_XYZ")
class TestClose:
    """Test close method of the client"""

    def test_close_empty(self):
        """Test client can close when no trial is reserved"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            client.close()

    def test_close_with_reserved(self):
        """Test client cannot be closed if trials are reserved."""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            with client.suggest() as trial:
                with pytest.raises(RuntimeError) as exc:
                    client.close()

                assert "There is still reserved trials" in str(exc.value)


@pytest.mark.usefixtures("version_XYZ")
class TestBroken:
    """Test handling of broken trials"""

    def test_broken_trial(self):
        """Test that broken trials are detected"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            with pytest.raises(RuntimeError):
                with client.suggest() as trial:
                    assert trial.status == "reserved"
                    raise RuntimeError("Dummy failure!")

            assert client._pacemakers == {}
            assert client.get_trial(trial).status == "broken"

    def test_interrupted_trial(self):
        """Test that interrupted trials are not set to broken"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            with pytest.raises(KeyboardInterrupt):
                with client.suggest() as trial:
                    assert trial.status == "reserved"
                    raise KeyboardInterrupt

            assert client._pacemakers == {}
            assert client.get_trial(trial).status == "interrupted"

    def test_completed_then_interrupted_trial(self):
        """Test that interrupted trials are not set to broken"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            with pytest.raises(KeyboardInterrupt):
                with client.suggest() as trial:
                    assert trial.status == "reserved"
                    assert trial.results == []
                    assert get_storage().get_trial(trial).objective is None
                    client.observe(
                        trial, [dict(name="objective", type="objective", value=101)]
                    )
                    assert get_storage().get_trial(trial).objective.value == 101
                    assert trial.status == "completed"
                    raise KeyboardInterrupt

            assert client._pacemakers == {}
            assert client.get_trial(trial).status == "completed"


@pytest.mark.usefixtures("version_XYZ")
class TestSuggest:
    """Tests for ExperimentClient.suggest"""

    def test_suggest(self, monkeypatch):
        """Verify that suggest reserved availabe trials."""
        mock_space_iterate(monkeypatch)
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.suggest()
            assert trial.status == "reserved"
            assert trial.params["x"] == 1

            assert len(experiment.fetch_trials()) == 5
            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()

    def test_suggest_new(self):
        """Verify that suggest can create, register and reserved new trials."""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            for _ in range(3):
                trial = client.suggest()
                assert trial.status == "reserved"
                assert len(experiment.fetch_trials()) == 5
                assert client._pacemakers[trial.id].is_alive()
                client._pacemakers[trial.id].stop()

            trial = client.suggest()
            assert trial.status == "reserved"
            assert trial.params["x"] == 57.57
            assert len(experiment.fetch_trials()) > 5

            assert client._pacemakers[trial.id].is_alive()
            for trial_id in list(client._pacemakers.keys()):
                client._pacemakers.pop(trial_id).stop()

    def test_suggest_race_condition(self, monkeypatch):
        """Verify that race conditions to register new trials is handled"""
        mock_space_iterate(monkeypatch)
        new_value = 50.0

        # algo will suggest once an already existing trial
        def amnesia(num=1):
            """Suggest a new value and then always suggest the same"""
            if amnesia.count == 0:
                value = [0]
            else:
                value = [new_value]

            amnesia.count += 1

            return [value]

        amnesia.count = 0

        with create_experiment(config, base_trial, statuses=["completed"]) as (
            cfg,
            experiment,
            client,
        ):

            monkeypatch.setattr(experiment.algorithms, "suggest", amnesia)

            assert len(experiment.fetch_trials()) == 1

            trial = client.suggest()
            assert trial.status == "reserved"
            assert trial.params["x"] == new_value
            assert amnesia.count == 2

            assert len(experiment.fetch_trials()) == 2
            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()

    def test_suggest_algo_opt_out(self, monkeypatch):
        """Verify that None is returned when algo cannot sample new trials (opting opt)"""

        def opt_out(num=1):
            """Never suggest a new trial"""
            return None

        monkeypatch.setattr(orion.core.config.worker, "max_idle_time", -1)

        with create_experiment(config, base_trial, statuses=["completed"]) as (
            cfg,
            experiment,
            client,
        ):

            monkeypatch.setattr(experiment.algorithms, "suggest", opt_out)

            assert len(experiment.fetch_trials()) == 1

            with pytest.raises(SampleTimeout):
                client.suggest()

    def test_suggest_is_done(self):
        """Verify that completed experiments cannot suggest new trials"""
        with create_experiment(config, base_trial, statuses=["completed"] * 10) as (
            cfg,
            experiment,
            client,
        ):

            assert len(experiment.fetch_trials()) == 10
            assert client.is_done

            with pytest.raises(CompletedExperiment):
                client.suggest()

    def test_suggest_is_done_context_manager(self):
        """Verify that context manager handles None"""
        with create_experiment(config, base_trial, statuses=["completed"] * 10) as (
            cfg,
            experiment,
            client,
        ):

            assert len(experiment.fetch_trials()) == 10
            assert client.is_done

            with pytest.raises(CompletedExperiment):
                client.suggest()

    def test_suggest_is_broken(self):
        """Verify that broken experiments cannot suggest new trials"""
        with create_experiment(config, base_trial, statuses=["broken"] * 10) as (
            cfg,
            experiment,
            client,
        ):

            assert len(experiment.fetch_trials()) == 10
            assert client.is_broken

            with pytest.raises(BrokenExperiment):
                client.suggest()

    def test_suggest_is_done_race_condition(self, monkeypatch):
        """Verify that inability to suggest because is_done becomes True during produce() is
        handled.
        """
        with create_experiment(config, base_trial, statuses=["completed"] * 5) as (
            cfg,
            experiment,
            client,
        ):

            def is_done(self):
                """Experiment is done"""
                return True

            def set_is_done():
                """Set is_done while algo is trying to suggest"""
                monkeypatch.setattr(experiment.__class__, "is_done", property(is_done))

            monkeypatch.setattr(client._producer, "produce", set_is_done)

            assert len(experiment.fetch_trials()) == 5
            assert not client.is_done

            with pytest.raises(CompletedExperiment):
                client.suggest()

            assert len(experiment.fetch_trials()) == 5
            assert client.is_done

    def test_suggest_is_broken_race_condition(self, monkeypatch):
        """Verify that experiments that gets broken during local algo.suggest gets properly
        handled
        """
        with create_experiment(config, base_trial, statuses=["broken"] * 1) as (
            cfg,
            experiment,
            client,
        ):

            def is_broken(self):
                """Experiment is broken"""
                return True

            def set_is_broken():
                """Set is_broken while algo is trying to suggest"""
                monkeypatch.setattr(
                    experiment.__class__, "is_broken", property(is_broken)
                )

            monkeypatch.setattr(client._producer, "produce", set_is_broken)

            assert len(experiment.fetch_trials()) == 1
            assert not client.is_broken

            with pytest.raises(BrokenExperiment):
                client.suggest()

            assert len(experiment.fetch_trials()) == 1
            assert client.is_broken

    def test_suggest_hierarchical_space(self):
        """Verify that suggest returns trial with proper hierarchical parameter."""
        exp_config = copy.deepcopy(config)
        exp_config["space"] = {
            "a": {"x": "uniform(0, 10, discrete=True)"},
            "b": {"y": "loguniform(1e-08, 1)", "z": "choices(['voici', 'voila', 2])"},
        }
        with create_experiment(
            exp_config=exp_config, trial_config=base_trial, statuses=[]
        ) as (cfg, experiment, client):
            trial = client.suggest()
            assert trial.status == "reserved"
            assert len(trial.params) == 2
            assert "x" in trial.params["a"]
            assert "y" in trial.params["b"]
            assert "z" in trial.params["b"]

            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()


@pytest.mark.usefixtures("version_XYZ")
class TestObserve:
    """Tests for ExperimentClient.observe"""

    def test_observe(self):
        """Verify that `observe()` will update the storage"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(**cfg.trials[1])
            assert trial.results == []
            client.reserve(trial)
            assert get_storage().get_trial(trial).objective is None
            client.observe(trial, [dict(name="objective", type="objective", value=101)])
            assert get_storage().get_trial(trial).objective.value == 101

    def test_observe_unreserved(self):
        """Verify that `observe()` will fail on non-reserved trials"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(**cfg.trials[1])
            with pytest.raises(RuntimeError) as exc:
                client.observe(
                    trial, [dict(name="objective", type="objective", value=101)]
                )

            assert "Trial {} had no pacemakers. Was it reserved properly?".format(
                trial.id
            ) == str(exc.value)

    def test_observe_dont_exist(self):
        """Verify that `observe()` will fail on non-registered trials"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(experiment="idontexist", params=cfg.trials[0]["params"])
            with pytest.raises(ValueError) as exc:
                client.observe(
                    trial, [dict(name="objective", type="objective", value=101)]
                )

            assert "Trial {} does not exist in database.".format(trial.id) == str(
                exc.value
            )
            assert client._pacemakers == {}

    def test_observe_bad_results(self):
        """Verify that bad results type is detected and ValueError is raised"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(**cfg.trials[1])
            client.reserve(trial)
            with pytest.raises(ValueError) as exc:
                client.observe(
                    trial, [dict(name="objective", type="bad bad bad", value=101)]
                )

            assert "Given type, bad bad bad, not one of: " in str(exc.value)
            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()

    def test_observe_race_condition(self):
        """Verify that race condition during `observe()` is detected and raised"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.get_trial(uid=cfg.trials[1]["_id"])
            client.reserve(trial)
            experiment.set_trial_status(trial, "interrupted")
            trial.status = "reserved"  # Let's pretend it is still reserved

            with pytest.raises(RuntimeError) as exc:
                client.observe(
                    trial, [dict(name="objective", type="objective", value=101)]
                )

            assert "Reservation for trial {} has been lost.".format(trial.id) == str(
                exc.value
            )
            assert client._pacemakers == {}

    def test_observe_under_with(self):
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            with client.suggest() as trial:
                assert trial.status == "reserved"
                assert trial.results == []
                assert get_storage().get_trial(trial).objective is None
                client.observe(
                    trial, [dict(name="objective", type="objective", value=101)]
                )
                assert get_storage().get_trial(trial).objective.value == 101
                assert trial.status == "completed"

            assert trial.status == "completed"  # Still completed after __exit__


@pytest.mark.usefixtures("version_XYZ")
class TestWorkon:
    """Tests for ExperimentClient.workon"""

    def test_workon(self):
        """Verify that workon processes properly"""

        def foo(x):
            return [dict(name="result", type="objective", value=x * 2)]

        with create_experiment(config, base_trial, statuses=[]) as (
            cfg,
            experiment,
            client,
        ):
            client.workon(foo, max_trials=5)
            assert len(experiment.fetch_trials_by_status("completed")) == 5
            assert client._pacemakers == {}

    def test_workon_partial(self):
        """Verify that partial is properly passed to the function"""

        def foo(x, y):
            return [dict(name="result", type="objective", value=x * 2 + y)]

        with create_experiment(config, base_trial, statuses=[]) as (
            cfg,
            experiment,
            client,
        ):
            client.workon(foo, max_trials=10, y=2)
            assert len(experiment.fetch_trials()) == 10
            assert client._pacemakers == {}

    def test_workon_partial_with_override(self):
        """Verify that partial is overriden by trial.params"""

        def foo(x, y):
            return [dict(name="result", type="objective", value=x * 2 + y)]

        ext_config = copy.deepcopy(config)
        ext_config["space"]["y"] = "uniform(0, 10)"

        with create_experiment(
            exp_config=ext_config, trial_config=base_trial, statuses=[]
        ) as (cfg, experiment, client):
            default_y = 2
            assert len(experiment.fetch_trials()) == 0
            client.workon(foo, max_trials=1, y=default_y)
            assert len(experiment.fetch_trials_by_status("completed")) == 1
            assert experiment.fetch_trials()[0].params["y"] != 2

    def test_workon_hierarchical_partial_with_override(self):
        """Verify that hierarchical partial is overriden by trial.params"""
        default_y = 2
        default_z = "voila"

        def foo(a, b):
            assert b["y"] != default_y
            assert b["z"] == default_z
            return [dict(name="result", type="objective", value=a["x"] * 2 + b["y"])]

        ext_config = copy.deepcopy(config)
        ext_config["space"] = {
            "a": {"x": "uniform(0, 10, discrete=True)"},
            "b": {"y": "loguniform(1e-08, 1)"},
        }

        with create_experiment(
            exp_config=ext_config, trial_config=base_trial, statuses=[]
        ) as (cfg, experiment, client):
            assert len(experiment.fetch_trials()) == 0
            client.workon(foo, max_trials=5, b={"y": default_y, "z": default_z})
            assert len(experiment.fetch_trials_by_status("completed")) == 5
            params = experiment.fetch_trials()[0].params
            assert len(params)
            assert "x" in params["a"]
            assert "y" in params["b"]

    def test_workon_max_trials(self):
        """Verify that workon stop when reaching max_trials"""

        def foo(x):
            return [dict(name="result", type="objective", value=x * 2)]

        with create_experiment(config, base_trial, statuses=[]) as (
            cfg,
            experiment,
            client,
        ):
            MAX_TRIALS = 5
            assert client.max_trials > MAX_TRIALS
            client.workon(foo, max_trials=MAX_TRIALS)
            assert len(experiment.fetch_trials_by_status("completed")) == MAX_TRIALS

    def test_workon_max_trials_resumed(self):
        """Verify that workon stop when reaching max_trials after resuming"""

        def foo(x):
            return [dict(name="result", type="objective", value=x * 2)]

        with create_experiment(
            config, base_trial, statuses=["completed", "completed"]
        ) as (
            cfg,
            experiment,
            client,
        ):
            MAX_TRIALS = 5
            assert client.max_trials > MAX_TRIALS
            assert len(experiment.fetch_trials_by_status("completed")) == 2
            client.workon(foo, max_trials=MAX_TRIALS)
            assert len(experiment.fetch_trials_by_status("completed")) == MAX_TRIALS

    def test_workon_max_trials_per_worker(self):
        """Verify that workon stop when reaching max_trials_per_worker"""

        def foo(x):
            return [dict(name="result", type="objective", value=x * 2)]

        with create_experiment(config, base_trial, statuses=[]) as (
            cfg,
            experiment,
            client,
        ):
            MAX_TRIALS = 5
            assert client.max_trials > MAX_TRIALS
            executed = client.workon(
                foo, max_trials=MAX_TRIALS, max_trials_per_worker=MAX_TRIALS - 1
            )
            assert executed == MAX_TRIALS - 1
            assert len(experiment.fetch_trials_by_status("completed")) == MAX_TRIALS - 1

    def test_workon_max_trials_per_worker_resumed(self):
        """Verify that workon stop when reaching max_trials_per_worker after resuming"""

        def foo(x):
            return [dict(name="result", type="objective", value=x * 2)]

        n_completed = 2
        statuses = ["completed"] * n_completed + ["new"]
        n_trials = len(statuses)

        with create_experiment(config, base_trial, statuses=statuses) as (
            cfg,
            experiment,
            client,
        ):
            MAX_TRIALS = 9
            assert client.max_trials > MAX_TRIALS
            assert len(experiment.fetch_trials_by_status("completed")) == n_completed
            executed = client.workon(
                foo, max_trials=MAX_TRIALS, max_trials_per_worker=2
            )
            assert executed == 2
            assert (
                len(experiment.fetch_trials_by_status("completed")) == 2 + n_completed
            )
            executed = client.workon(
                foo, max_trials=MAX_TRIALS, max_trials_per_worker=3
            )
            assert executed == 3
            assert (
                len(experiment.fetch_trials_by_status("completed"))
                == 3 + 2 + n_completed
            )

    def test_workon_exp_max_broken_before_worker_max_broken(self):
        """Verify that workon stop when reaching exp.max_broken"""

        def foo(x):
            raise RuntimeError()

        MAX_TRIALS = 5
        MAX_BROKEN = 20
        test_config = copy.deepcopy(config)
        test_config["max_broken"] = MAX_BROKEN // 2

        with create_experiment(test_config, base_trial, statuses=[]) as (
            cfg,
            experiment,
            client,
        ):
            with pytest.raises(BrokenExperiment):
                client.workon(foo, max_trials=MAX_TRIALS, max_broken=MAX_BROKEN)
            n_broken_trials = len(experiment.fetch_trials_by_status("broken"))
            n_trials = len(experiment.fetch_trials())
            assert n_broken_trials == MAX_BROKEN // 2
            assert n_trials - n_broken_trials < MAX_TRIALS

    def test_workon_max_broken_all_broken(self):
        """Verify that workon stop when reaching worker's max_broken"""

        def foo(x):
            raise RuntimeError()

        MAX_TRIALS = 5
        MAX_BROKEN = 10

        test_config = copy.deepcopy(config)
        test_config["max_broken"] = MAX_BROKEN * 2

        with create_experiment(test_config, base_trial, statuses=[]) as (
            cfg,
            experiment,
            client,
        ):
            with pytest.raises(BrokenExperiment):
                client.workon(foo, max_trials=MAX_TRIALS, max_broken=MAX_BROKEN)
            n_broken_trials = len(experiment.fetch_trials_by_status("broken"))
            n_trials = len(experiment.fetch_trials())
            assert n_broken_trials == MAX_BROKEN
            assert n_trials - n_broken_trials < MAX_TRIALS

    def test_workon_max_trials_before_max_broken(self):
        """Verify that workon stop when reaching max_trials before max_broken"""

        with create_experiment(config, base_trial, statuses=[]) as (
            cfg,
            experiment,
            client,
        ):

            def foo(x):
                if len(client.fetch_trials()) < 5:
                    raise RuntimeError()

                return [dict(name="result", type="objective", value=x * 2)]

            MAX_TRIALS = 5
            MAX_BROKEN = 10
            assert client.max_trials > MAX_TRIALS
            client.workon(foo, max_trials=MAX_TRIALS, max_broken=MAX_BROKEN)
            n_broken_trials = len(experiment.fetch_trials_by_status("broken"))
            n_trials = len(experiment.fetch_trials())
            assert n_broken_trials < MAX_BROKEN
            assert n_trials - n_broken_trials == MAX_TRIALS

    def test_workon_trial_arg(self):
        """Verify that workon pass trial when trial_arg is defined"""

        def foo(x, my_trial_arg_name):
            assert isinstance(my_trial_arg_name, Trial)
            assert my_trial_arg_name.params["x"] == x
            return [dict(name="result", type="objective", value=x * 2)]

        with create_experiment(config, base_trial, statuses=[]) as (
            cfg,
            experiment,
            client,
        ):
            client.workon(foo, max_trials=5, trial_arg="my_trial_arg_name")
            assert len(experiment.fetch_trials()) == 5

    def test_workon_on_error_ignore(self):
        """Verify that workon on_error callback ignores some errors correctly"""

        def on_error(client, trial, error, worker_broken_trials):
            assert on_error.counter == worker_broken_trials
            if isinstance(error, (IndexError, IOError, AttributeError)):
                client.release(trial, "cancelled")
                return False

            on_error.counter += 1
            return True

        on_error.counter = 0

        errors = [
            IndexError,
            ValueError,
            IOError,
            NotImplementedError,
            AttributeError,
            ImportError,
        ]

        def foo(x):
            if errors:
                raise errors.pop()()

            return [dict(name="result", type="objective", value=x * 2)]

        MAX_TRIALS = 5
        MAX_BROKEN = len(errors) + 1

        test_config = copy.deepcopy(config)
        test_config["max_broken"] = MAX_BROKEN * 2

        with create_experiment(test_config, base_trial, statuses=[]) as (
            cfg,
            experiment,
            client,
        ):

            client.workon(foo, max_trials=MAX_TRIALS, max_broken=MAX_BROKEN)
            n_broken_trials = len(experiment.fetch_trials_by_status("broken"))
            n_trials = len(experiment.fetch_trials())
            assert n_broken_trials == MAX_BROKEN - 1
            assert n_trials - n_broken_trials == MAX_TRIALS

    def test_workon_on_error_raise(self):
        """Verify that workon on_error callback can raise and stop iteration"""

        def on_error(client, trial, error, worker_broken_trials):
            raise error

        def foo(x):
            raise NotImplementedError("Do not ignore this!")

        with create_experiment(config, base_trial, statuses=[]) as (
            cfg,
            experiment,
            client,
        ):
            with pytest.raises(NotImplementedError) as exc:
                client.workon(foo, max_trials=5, max_broken=5, on_error=on_error)

            assert exc.match("Do not ignore this!")

    def test_parallel_workers(self, monkeypatch):
        """Test parallel execution with joblib"""

        def foo(x):
            return [dict(name="result", type="objective", value=x * 2)]

        def optimize(*args, **kwargs):
            optimize.count += 1
            return 1

        with create_experiment(exp_config=config, trial_config={}, statuses=[]) as (
            cfg,
            experiment,
            client,
        ):

            monkeypatch.setattr(client, "_optimize", optimize)
            optimize.count = 0
            with client.tmp_executor("joblib", n_workers=5, backend="threading"):
                client.workon(foo, max_trials=5, n_workers=2)

            assert optimize.count == 2
            optimize.count = 0
            with client.tmp_executor("joblib", n_workers=5, backend="threading"):
                client.workon(foo, max_trials=5, n_workers=3)
            assert optimize.count == 3



"""
Benchmark Assessments definition
=================================
"""

from .averagerank import AverageRank
from .averageresult import AverageResult
from .base import BaseAssess

__all__ = ["BaseAssess", "AverageRank", "AverageResult"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a functional test for branching."""

import os

import pytest

import orion.core.cli
import orion.core.io.experiment_builder as experiment_builder
from orion.storage.base import get_storage


def execute(command, assert_code=0):
    """Execute orion command and return returncode"""
    returncode = orion.core.cli.main(command.split(" "))
    assert returncode == assert_code


@pytest.fixture
def init_full_x(setup_pickleddb_database, monkeypatch):
    """Init original experiment"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    name = "full_x"
    orion.core.cli.main(
        (
            "hunt --init-only -n {name} --config orion_config.yaml ./black_box.py "
            "-x~uniform(-10,10)"
        )
        .format(name=name)
        .split(" ")
    )
    orion.core.cli.main("insert -n {name} script -x=0".format(name=name).split(" "))


@pytest.fixture
def init_full_x_full_y(init_full_x):
    """Add y dimension to original"""
    name = "full_x"
    branch = "full_x_full_y"
    orion.core.cli.main(
        (
            "hunt --init-only -n {branch} --branch-from {name} --cli-change-type noeffect "
            "./black_box_with_y.py "
            "-x~uniform(-10,10) "
            "-y~+uniform(-10,10,default_value=1)"
        )
        .format(name=name, branch=branch)
        .split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=1 -y=1".format(branch=branch).split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=-1 -y=1".format(branch=branch).split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=1 -y=-1".format(branch=branch).split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=-1 -y=-1".format(branch=branch).split(" ")
    )


@pytest.fixture
def init_half_x_full_y(init_full_x_full_y):
    """Change x's prior to full x and full y experiment"""
    name = "full_x_full_y"
    branch = "half_x_full_y"
    orion.core.cli.main(
        (
            "hunt --init-only -n {branch} --branch-from {name} ./black_box_with_y.py "
            "-x~+uniform(0,10) "
            "-y~uniform(-10,10,default_value=1)"
        )
        .format(name=name, branch=branch)
        .split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=2 -y=2".format(branch=branch).split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=2 -y=-2".format(branch=branch).split(" ")
    )


@pytest.fixture
def init_full_x_half_y(init_full_x_full_y):
    """Change y's prior to full x and full y experiment"""
    name = "full_x_full_y"
    branch = "full_x_half_y"
    orion.core.cli.main(
        (
            "hunt --init-only -n {branch} --branch-from {name} ./black_box_with_y.py "
            "-x~uniform(-10,10) "
            "-y~+uniform(0,10,default_value=1)"
        )
        .format(name=name, branch=branch)
        .split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=3 -y=3".format(branch=branch).split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=-3 -y=3".format(branch=branch).split(" ")
    )


@pytest.fixture
def init_full_x_rename_y_z(init_full_x_full_y):
    """Rename y from full x full y to z"""
    name = "full_x_full_y"
    branch = "full_x_rename_y_z"
    orion.core.cli.main(
        (
            "hunt --init-only -n {branch} --branch-from {name} --cli-change-type noeffect "
            "./black_box_with_z.py -x~uniform(-10,10) -y~>z -z~uniform(-10,10,default_value=1)"
        )
        .format(name=name, branch=branch)
        .split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=4 -z=4".format(branch=branch).split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=-4 -z=4".format(branch=branch).split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=4 -z=-4".format(branch=branch).split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=-4 -z=-4".format(branch=branch).split(" ")
    )


@pytest.fixture
def init_full_x_rename_half_y_half_z(init_full_x_half_y):
    """Rename y from full x half y to z"""
    name = "full_x_half_y"
    branch = "full_x_rename_half_y_half_z"
    orion.core.cli.main(
        (
            "hunt --init-only -n {branch} --branch-from {name} --cli-change-type noeffect "
            "./black_box_with_z.py -x~uniform(-10,10) -y~>z -z~uniform(0,10,default_value=1)"
        )
        .format(name=name, branch=branch)
        .split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=5 -z=5".format(branch=branch).split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=-5 -z=5".format(branch=branch).split(" ")
    )


@pytest.fixture
def init_full_x_rename_half_y_full_z(init_full_x_half_y):
    """Rename y from full x half y to full z (rename + changed prior)"""
    name = "full_x_half_y"
    branch = "full_x_rename_half_y_full_z"
    orion.core.cli.main(
        (
            "hunt --init-only -n {branch} --branch-from {name} --cli-change-type noeffect "
            "./black_box_with_z.py "
            "-x~uniform(-10,10) -y~>z "
            "-z~+uniform(-10,10,default_value=1)"
        )
        .format(name=name, branch=branch)
        .split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=6 -z=6".format(branch=branch).split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=-6 -z=6".format(branch=branch).split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=6 -z=-6".format(branch=branch).split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=-6 -z=-6".format(branch=branch).split(" ")
    )


@pytest.fixture
def init_full_x_remove_y(init_full_x_full_y):
    """Remove y from full x full y"""
    name = "full_x_full_y"
    branch = "full_x_remove_y"
    orion.core.cli.main(
        (
            "hunt --init-only -n {branch} --branch-from {name} --cli-change-type noeffect "
            "./black_box.py "
            "-x~uniform(-10,10) -y~-"
        )
        .format(name=name, branch=branch)
        .split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=7".format(branch=branch).split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=-7".format(branch=branch).split(" ")
    )


@pytest.fixture
def init_full_x_full_y_add_z_remove_y(init_full_x_full_y):
    """Remove y from full x full y and add z"""
    name = "full_x_full_y"
    branch = "full_x_full_z_remove_y"
    orion.core.cli.main(
        (
            "hunt --init-only -n {branch} --branch-from {name} --cli-change-type noeffect "
            "./black_box.py -x~uniform(-10,10) "
            "-z~uniform(-20,10,default_value=0)"
        )
        .format(name=name, branch=branch)
        .split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=7 -z=2".format(branch=branch).split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=-7 -z=2".format(branch=branch).split(" ")
    )


@pytest.fixture
def init_full_x_remove_z(init_full_x_rename_y_z):
    """Remove z from full x full z"""
    name = "full_x_rename_y_z"
    branch = "full_x_remove_z"
    orion.core.cli.main(
        (
            "hunt --init-only -n {branch} --branch-from {name} --cli-change-type noeffect "
            "./black_box.py "
            "-x~uniform(-10,10) -z~-"
        )
        .format(name=name, branch=branch)
        .split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=8".format(branch=branch).split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=-8".format(branch=branch).split(" ")
    )


@pytest.fixture
def init_full_x_remove_z_default_4(init_full_x_rename_y_z):
    """Remove z from full x full z and give a default value of 4"""
    name = "full_x_rename_y_z"
    branch = "full_x_remove_z_default_4"
    orion.core.cli.main(
        (
            "hunt --init-only -n {branch} --branch-from {name} --cli-change-type noeffect "
            "./black_box.py "
            "-x~uniform(-10,10) -z~-4"
        )
        .format(name=name, branch=branch)
        .split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=9".format(branch=branch).split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=-9".format(branch=branch).split(" ")
    )


@pytest.fixture
def init_full_x_new_algo(init_full_x):
    """Remove z from full x full z and give a default value of 4"""
    name = "full_x"
    branch = "full_x_new_algo"
    orion.core.cli.main(
        (
            "hunt --init-only -n {branch} --branch-from {name} "
            "--algorithm-change --config new_algo_config.yaml "
            "./black_box.py -x~uniform(-10,10)"
        )
        .format(name=name, branch=branch)
        .split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=1.1".format(branch=branch).split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=-1.1".format(branch=branch).split(" ")
    )


@pytest.fixture
def init_full_x_new_cli(init_full_x):
    """Remove z from full x full z and give a default value of 4"""
    name = "full_x"
    branch = "full_x_new_cli"
    orion.core.cli.main(
        (
            "hunt --init-only -n {branch} --branch-from {name} --cli-change-type noeffect "
            "./black_box_new.py -x~uniform(-10,10) --a-new argument"
        )
        .format(name=name, branch=branch)
        .split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=1.2".format(branch=branch).split(" ")
    )
    orion.core.cli.main(
        "insert -n {branch} script -x=-1.2".format(branch=branch).split(" ")
    )


@pytest.fixture
def init_full_x_ignore_cli(init_full_x):
    """Use the --non-monitored-arguments argument"""
    name = "full_x_with_new_opt"
    orion.core.cli.main(
        (
            "hunt --init-only -n {name} --config orion_config.yaml ./black_box_new.py "
            "-x~uniform(-10,10)"
        )
        .format(name=name)
        .split(" ")
    )
    orion.core.cli.main("insert -n {name} script -x=0".format(name=name).split(" "))

    orion.core.cli.main(
        (
            "hunt --init-only -n {name} --non-monitored-arguments a-new "
            "--config orion_config.yaml ./black_box_new.py "
            "-x~uniform(-10,10) --a-new argument"
        )
        .format(name=name)
        .split(" ")
    )
    orion.core.cli.main("insert -n {name} script -x=1.2".format(name=name).split(" "))
    orion.core.cli.main("insert -n {name} script -x=-1.2".format(name=name).split(" "))


@pytest.fixture
def init_entire(
    init_half_x_full_y,  # 1.1.1
    init_full_x_rename_half_y_half_z,  # 1.1.2.1
    init_full_x_rename_half_y_full_z,  # 1.1.2.2
    init_full_x_remove_y,  # 1.1.4
    init_full_x_remove_z,  # 1.1.3.1
    init_full_x_remove_z_default_4,
):  # 1.1.3.2
    """Initialize all experiments"""
    pass


def get_name_value_pairs(trials):
    """Turn parameters into pairs for easy comparisions"""
    pairs = []
    for trial in trials:
        pairs.append([])
        for param in trial._params:
            pairs[-1].append((param.name, param.value))

        pairs[-1] = tuple(pairs[-1])

    return tuple(pairs)


def test_init(init_full_x):
    """Test if original experiment contains trial 0"""
    experiment = experiment_builder.load(name="full_x")

    assert experiment.refers["adapter"].configuration == []

    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == ((("/x", 0),),)


def test_full_x_full_y(init_full_x_full_y):
    """Test if full x full y is properly initialized and can fetch original trial"""
    experiment = experiment_builder.load(name="full_x_full_y")

    assert experiment.refers["adapter"].configuration == [
        {"change_type": "noeffect", "of_type": "commandlinechange"},
        {
            "of_type": "dimensionaddition",
            "param": {"name": "/y", "type": "real", "value": 1},
        },
    ]

    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == (
        (("/x", 1), ("/y", 1)),
        (("/x", -1), ("/y", 1)),
        (("/x", 1), ("/y", -1)),
        (("/x", -1), ("/y", -1)),
    )

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    assert pairs == (
        (("/x", 0), ("/y", 1)),
        (("/x", 1), ("/y", 1)),
        (("/x", -1), ("/y", 1)),
        (("/x", 1), ("/y", -1)),
        (("/x", -1), ("/y", -1)),
    )


def test_half_x_full_y(init_half_x_full_y):
    """Test if half x full y is properly initialized and can fetch from its 2 parents"""
    experiment = experiment_builder.load(name="half_x_full_y")

    assert experiment.refers["adapter"].configuration == [
        {
            "of_type": "dimensionpriorchange",
            "name": "/x",
            "old_prior": "uniform(-10, 10)",
            "new_prior": "uniform(0, 10)",
        }
    ]

    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == ((("/x", 2), ("/y", 2)), (("/x", 2), ("/y", -2)))

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    assert pairs == (
        (("/x", 0), ("/y", 1)),
        (("/x", 1), ("/y", 1)),
        (("/x", 1), ("/y", -1)),
        (("/x", 2), ("/y", 2)),
        (("/x", 2), ("/y", -2)),
    )


def test_full_x_half_y(init_full_x_half_y):
    """Test if full x half y is properly initialized and can fetch from its 2 parents"""
    experiment = experiment_builder.load(name="full_x_half_y")

    assert experiment.refers["adapter"].configuration == [
        {
            "of_type": "dimensionpriorchange",
            "name": "/y",
            "old_prior": "uniform(-10, 10, default_value=1)",
            "new_prior": "uniform(0, 10, default_value=1)",
        }
    ]

    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == ((("/x", 3), ("/y", 3)), (("/x", -3), ("/y", 3)))

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    assert pairs == (
        (("/x", 0), ("/y", 1)),
        (("/x", 1), ("/y", 1)),
        (("/x", -1), ("/y", 1)),
        (("/x", 3), ("/y", 3)),
        (("/x", -3), ("/y", 3)),
    )


def test_full_x_rename_y_z(init_full_x_rename_y_z):
    """Test if full x full z is properly initialized and can fetch from its 2 parents"""
    experiment = experiment_builder.load(name="full_x_rename_y_z")

    assert experiment.refers["adapter"].configuration == [
        {"of_type": "commandlinechange", "change_type": "noeffect"},
        {"of_type": "dimensionrenaming", "old_name": "/y", "new_name": "/z"},
    ]

    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == (
        (("/x", 4), ("/z", 4)),
        (("/x", -4), ("/z", 4)),
        (("/x", 4), ("/z", -4)),
        (("/x", -4), ("/z", -4)),
    )

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    assert pairs == (
        (("/x", 0), ("/z", 1)),
        (("/x", 1), ("/z", 1)),
        (("/x", -1), ("/z", 1)),
        (("/x", 1), ("/z", -1)),
        (("/x", -1), ("/z", -1)),
        (("/x", 4), ("/z", 4)),
        (("/x", -4), ("/z", 4)),
        (("/x", 4), ("/z", -4)),
        (("/x", -4), ("/z", -4)),
    )


def test_full_x_rename_half_y_half_z(init_full_x_rename_half_y_half_z):
    """Test if full x half z is properly initialized and can fetch from its 3 parents"""
    experiment = experiment_builder.load(name="full_x_rename_half_y_half_z")

    assert experiment.refers["adapter"].configuration == [
        {"of_type": "commandlinechange", "change_type": "noeffect"},
        {"of_type": "dimensionrenaming", "old_name": "/y", "new_name": "/z"},
    ]

    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == ((("/x", 5), ("/z", 5)), (("/x", -5), ("/z", 5)))

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    assert pairs == (
        (("/x", 0), ("/z", 1)),
        (("/x", 1), ("/z", 1)),
        (("/x", -1), ("/z", 1)),
        (("/x", 3), ("/z", 3)),
        (("/x", -3), ("/z", 3)),
        (("/x", 5), ("/z", 5)),
        (("/x", -5), ("/z", 5)),
    )


def test_full_x_rename_half_y_full_z(init_full_x_rename_half_y_full_z):
    """Test if full x half->full z is properly initialized and can fetch from its 3 parents"""
    experiment = experiment_builder.load(name="full_x_rename_half_y_full_z")

    assert experiment.refers["adapter"].configuration == [
        {"of_type": "commandlinechange", "change_type": "noeffect"},
        {"of_type": "dimensionrenaming", "old_name": "/y", "new_name": "/z"},
        {
            "of_type": "dimensionpriorchange",
            "name": "/z",
            "old_prior": "uniform(0, 10, default_value=1)",
            "new_prior": "uniform(-10, 10, default_value=1)",
        },
    ]

    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == (
        (("/x", 6), ("/z", 6)),
        (("/x", -6), ("/z", 6)),
        (("/x", 6), ("/z", -6)),
        (("/x", -6), ("/z", -6)),
    )

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    assert pairs == (
        (("/x", 0), ("/z", 1)),
        (("/x", 1), ("/z", 1)),
        (("/x", -1), ("/z", 1)),
        (("/x", 3), ("/z", 3)),
        (("/x", -3), ("/z", 3)),
        (("/x", 6), ("/z", 6)),
        (("/x", -6), ("/z", 6)),
        (("/x", 6), ("/z", -6)),
        (("/x", -6), ("/z", -6)),
    )


def test_full_x_remove_y(init_full_x_remove_y):
    """Test if full x removed y is properly initialized and can fetch from its 2 parents"""
    experiment = experiment_builder.load(name="full_x_remove_y")

    assert experiment.refers["adapter"].configuration == [
        {"of_type": "commandlinechange", "change_type": "noeffect"},
        {
            "of_type": "dimensiondeletion",
            "param": {"name": "/y", "type": "real", "value": 1},
        },
    ]

    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == ((("/x", 7),), (("/x", -7),))

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    assert pairs == (
        (("/x", 0),),
        (("/x", 1),),
        (("/x", -1),),
        (("/x", 7),),
        (("/x", -7),),
    )


def test_full_x_full_y_add_z_remove_y(init_full_x_full_y_add_z_remove_y):
    """Test that if z is added and y removed at the same time, both are correctly detected"""
    experiment = experiment_builder.load(name="full_x_full_z_remove_y")

    assert experiment.refers["adapter"].configuration == [
        {"of_type": "commandlinechange", "change_type": "noeffect"},
        {
            "of_type": "dimensiondeletion",
            "param": {"name": "/y", "type": "real", "value": 1},
        },
        {
            "of_type": "dimensionaddition",
            "param": {"name": "/z", "type": "real", "value": 0},
        },
    ]

    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == ((("/x", 7), ("/z", 2)), (("/x", -7), ("/z", 2)))

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    assert pairs == (
        (("/x", 0), ("/z", 0)),
        (("/x", 1), ("/z", 0)),
        (("/x", -1), ("/z", 0)),
        (("/x", 7), ("/z", 2)),
        (("/x", -7), ("/z", 2)),
    )


def test_full_x_remove_z(init_full_x_remove_z):
    """Test if full x removed z is properly initialized and can fetch from 2 of its 3 parents"""
    experiment = experiment_builder.load(name="full_x_remove_z")

    assert experiment.refers["adapter"].configuration == [
        {"change_type": "noeffect", "of_type": "commandlinechange"},
        {
            "of_type": "dimensiondeletion",
            "param": {"name": "/z", "type": "real", "value": 1},
        },
    ]

    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == ((("/x", 8),), (("/x", -8),))

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    # Note that full_x_rename_y_z are filtered out because default_value=1
    assert pairs == (
        (("/x", 0),),
        (("/x", 1),),
        (("/x", -1),),
        (("/x", 8),),
        (("/x", -8),),
    )


def test_full_x_remove_z_default_4(init_full_x_remove_z_default_4):
    """Test if full x removed z  (default 4) is properly initialized and can fetch
    from 1 of its 3 parents
    """
    experiment = experiment_builder.load(name="full_x_remove_z_default_4")

    assert experiment.refers["adapter"].configuration == [
        {"change_type": "noeffect", "of_type": "commandlinechange"},
        {
            "of_type": "dimensiondeletion",
            "param": {"name": "/z", "type": "real", "value": 4.0},
        },
    ]

    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == ((("/x", 9),), (("/x", -9),))

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    # Note that full_x and full_x_full_y are filtered out because default_value=4
    assert pairs == ((("/x", 4),), (("/x", -4),), (("/x", 9),), (("/x", -9),))


def test_entire_full_x_full_y(init_entire):
    """Test if full x full y can fetch from its parent and all children"""
    experiment = experiment_builder.load(name="full_x_full_y")

    assert experiment.refers["adapter"].configuration == [
        {"change_type": "noeffect", "of_type": "commandlinechange"},
        {
            "of_type": "dimensionaddition",
            "param": {"name": "/y", "type": "real", "value": 1},
        },
    ]

    pairs = get_name_value_pairs(experiment.fetch_trials())
    assert pairs == (
        (("/x", 1), ("/y", 1)),
        (("/x", -1), ("/y", 1)),
        (("/x", 1), ("/y", -1)),
        (("/x", -1), ("/y", -1)),
    )

    pairs = get_name_value_pairs(experiment.fetch_trials(with_evc_tree=True))
    assert set(pairs) == set(
        (
            (("/x", 0), ("/y", 1)),
            # full_x_full_y
            (("/x", 1), ("/y", 1)),
            (("/x", -1), ("/y", 1)),
            (("/x", 1), ("/y", -1)),
            (("/x", -1), ("/y", -1)),
            # half_x_full_y
            (("/x", 2), ("/y", 2)),
            (("/x", 2), ("/y", -2)),
            # full_x_half_y
            (("/x", 3), ("/y", 3)),
            (("/x", -3), ("/y", 3)),
            # full_x_rename_y_z
            (("/x", 4), ("/y", 4)),
            (("/x", -4), ("/y", 4)),
            (("/x", 4), ("/y", -4)),
            (("/x", -4), ("/y", -4)),
            # full_x_rename_half_y_half_z
            (("/x", 5), ("/y", 5)),
            (("/x", -5), ("/y", 5)),
            # full_x_rename_half_y_full_z
            (("/x", 6), ("/y", 6)),
            (("/x", -6), ("/y", 6)),
            # full_x_remove_y
            (("/x", 7), ("/y", 1)),
            (("/x", -7), ("/y", 1)),
            # full_x_remove_z
            (("/x", 8), ("/y", 1)),
            (("/x", -8), ("/y", 1)),
            # full_x_remove_z_default_4
            (("/x", 9), ("/y", 4)),
            (("/x", -9), ("/y", 4)),
        )
    )


def test_run_entire_full_x_full_y(init_entire):
    """Test if branched experiment can be executed without triggering a branching event again"""
    experiment = experiment_builder.load(name="full_x_full_y")

    assert experiment.refers["adapter"].configuration == [
        {"change_type": "noeffect", "of_type": "commandlinechange"},
        {
            "of_type": "dimensionaddition",
            "param": {"name": "/y", "type": "real", "value": 1},
        },
    ]

    assert len(experiment.fetch_trials(with_evc_tree=True)) == 23
    assert len(experiment.fetch_trials()) == 4

    orion.core.cli.main(
        (
            "-vv hunt --max-trials 20 --pool-size 1 -n full_x_full_y "
            "./black_box_with_y.py "
            "-x~uniform(-10,10) "
            "-y~uniform(-10,10,default_value=1)"
        ).split(" ")
    )

    assert len(experiment.fetch_trials(with_evc_tree=True)) == 39
    assert len(experiment.fetch_trials()) == 20


def test_run_entire_full_x_full_y_no_args(init_entire):
    """Test if branched experiment can be executed without script arguments"""
    experiment = experiment_builder.load(name="full_x_full_y")

    assert len(experiment.fetch_trials(with_evc_tree=True)) == 23
    assert len(experiment.fetch_trials()) == 4

    orion.core.cli.main(
        ("-vv hunt --max-trials 20 --pool-size 1 -n full_x_full_y").split(" ")
    )

    assert len(experiment.fetch_trials(with_evc_tree=True)) == 39
    assert len(experiment.fetch_trials()) == 20


def test_new_algo(init_full_x_new_algo):
    """Test that new algo conflict is automatically resolved"""
    experiment = experiment_builder.load(name="full_x_new_algo")

    assert experiment.refers["adapter"].configuration == [
        {"of_type": "algorithmchange"}
    ]

    assert len(experiment.fetch_trials(with_evc_tree=True)) == 3
    assert len(experiment.fetch_trials()) == 2

    orion.core.cli.main(
        ("-vv hunt --max-trials 20 --pool-size 1 -n full_x_new_algo").split(" ")
    )

    assert len(experiment.fetch_trials(with_evc_tree=True)) == 21
    assert len(experiment.fetch_trials()) == 20


def test_new_algo_not_resolved(init_full_x, capsys):
    """Test that new algo conflict is not automatically resolved"""
    name = "full_x"
    branch = "full_x_new_algo"
    error_code = orion.core.cli.main(
        (
            "hunt --init-only -n {branch} --branch-from {name} --config new_algo_config.yaml "
            "--manual-resolution ./black_box.py -x~uniform(-10,10)"
        )
        .format(name=name, branch=branch)
        .split(" ")
    )
    assert error_code == 1

    captured = capsys.readouterr()
    assert captured.out == ""
    assert "Configuration is different and generates a branching event" in captured.err
    assert "gradient_descent" in captured.err


def test_ignore_cli(init_full_x_ignore_cli):
    """Test that a non-monitored parameter conflict is not generating a child"""
    name = "full_x"
    orion.core.cli.main(
        (
            "hunt --init-only -n {name} --non-monitored-arguments a-new "
            "--manual-resolution ./black_box.py -x~uniform(-10,10)"
        )
        .format(name=name)
        .split(" ")
    )


@pytest.mark.usefixtures("init_full_x", "mock_infer_versioning_metadata")
def test_new_code_triggers_code_conflict(capsys):
    """Test that a different git hash is generating a child"""
    name = "full_x"
    error_code = orion.core.cli.main(
        (
            "hunt --init-only -n {name} "
            "--manual-resolution ./black_box.py -x~uniform(-10,10)"
        )
        .format(name=name)
        .split(" ")
    )
    assert error_code == 1

    captured = capsys.readouterr()
    assert captured.out == ""
    assert "Configuration is different and generates a branching event" in captured.err
    assert "--code-change-type" in captured.err


@pytest.mark.usefixtures("init_full_x", "mock_infer_versioning_metadata")
def test_new_code_triggers_code_conflict_with_name_only(capsys):
    """Test that a different git hash is generating a child, even if cmdline is not passed"""
    name = "full_x"
    error_code = orion.core.cli.main(
        ("hunt --init-only -n {name} " "--manual-resolution")
        .format(name=name)
        .split(" ")
    )
    assert error_code == 1

    captured = capsys.readouterr()
    assert captured.out == ""
    assert "Configuration is different and generates a branching event" in captured.err
    assert "--code-change-type" in captured.err


@pytest.mark.usefixtures("init_full_x", "mock_infer_versioning_metadata")
def test_new_code_ignores_code_conflict():
    """Test that a different git hash is *not* generating a child if --ignore-code-changes"""
    name = "full_x"
    # Let it run for 2 trials to test consumer._validate_code_version too.
    error_code = orion.core.cli.main(
        (
            "hunt --worker-max-trials 2 -n {name} --ignore-code-changes "
            "--manual-resolution ./black_box.py -x~uniform(-10,10)"
        )
        .format(name=name)
        .split(" ")
    )
    assert error_code == 0


@pytest.mark.usefixtures("init_full_x", "version_XYZ")
def test_new_orion_version_triggers_conflict(capsys):
    """Test that a different git hash is generating a child"""
    name = "full_x"
    error_code = orion.core.cli.main(
        ("hunt --init-only -n {name} --manual-resolution").format(name=name).split(" ")
    )
    assert error_code == 1

    captured = capsys.readouterr()
    assert captured.out == ""
    assert "Configuration is different and generates a branching event" in captured.err
    assert "XYZ" in captured.err


def test_new_cli(init_full_x_new_cli):
    """Test that new cli conflict is automatically resolved"""
    experiment = experiment_builder.load(name="full_x_new_cli")

    assert experiment.refers["adapter"].configuration == [
        {"change_type": "noeffect", "of_type": "commandlinechange"}
    ]

    assert len(experiment.fetch_trials(with_evc_tree=True)) == 3
    assert len(experiment.fetch_trials()) == 2

    orion.core.cli.main(
        ("-vv hunt --max-trials 20 --pool-size 1 -n full_x_new_cli").split(" ")
    )

    assert len(experiment.fetch_trials(with_evc_tree=True)) == 21
    assert len(experiment.fetch_trials()) == 20


@pytest.mark.usefixtures("init_full_x")
def test_no_cli_no_branching():
    """Test that no branching occurs when using same code and not passing cmdline"""
    name = "full_x"
    error_code = orion.core.cli.main(
        ("hunt --init-only -n {name} " "--manual-resolution")
        .format(name=name)
        .split(" ")
    )
    assert error_code == 0


def test_auto_resolution_does_resolve(init_full_x_full_y, monkeypatch):
    """Test that auto-resolution does resolve all conflicts"""
    # Patch cmdloop to avoid autoresolution's prompt
    monkeypatch.setattr("sys.__stdin__.isatty", lambda: True)

    name = "full_x_full_y"
    branch = "half_x_no_y_new_w"
    # If autoresolution was not succesfull, this to fail with a sys.exit without registering the
    # experiment
    orion.core.cli.main(
        (
            "hunt --init-only -n {branch} --branch-from {name} ./black_box_with_y.py "
            "-x~uniform(0,10) "
            "-w~choices(['a','b'])"
        )
        .format(name=name, branch=branch)
        .split(" ")
    )

    experiment = experiment_builder.load(name=branch)

    assert experiment.refers["adapter"].configuration == [
        {
            "of_type": "dimensionpriorchange",
            "name": "/x",
            "new_prior": "uniform(0, 10)",
            "old_prior": "uniform(-10, 10)",
        },
        {
            "of_type": "dimensiondeletion",
            "param": {"name": "/y", "type": "real", "value": 1},
        },
        {
            "of_type": "dimensionaddition",
            "param": {"name": "/w", "type": "categorical", "value": None},
        },
    ]


def test_auto_resolution_with_fidelity(init_full_x_full_y, monkeypatch):
    """Test that auto-resolution does resolve all conflicts including new fidelity"""
    # Patch cmdloop to avoid autoresolution's prompt
    monkeypatch.setattr("sys.__stdin__.isatty", lambda: True)

    name = "full_x_full_y"
    branch = "half_x_no_y_new_w"
    # If autoresolution was not succesfull, this to fail with a sys.exit without registering the
    # experiment
    orion.core.cli.main(
        (
            "hunt --init-only -n {branch} --branch-from {name} ./black_box_with_y.py "
            "-x~uniform(0,10) "
            "-w~fidelity(1,10)"
        )
        .format(name=name, branch=branch)
        .split(" ")
    )

    experiment = experiment_builder.load(name=branch)

    assert experiment.refers["adapter"].configuration == [
        {
            "of_type": "dimensionpriorchange",
            "name": "/x",
            "new_prior": "uniform(0, 10)",
            "old_prior": "uniform(-10, 10)",
        },
        {
            "of_type": "dimensiondeletion",
            "param": {"name": "/y", "type": "real", "value": 1},
        },
        {
            "of_type": "dimensionaddition",
            "param": {"name": "/w", "type": "fidelity", "value": 10},
        },
    ]


def test_init_w_version_from_parent_w_children(
    setup_pickleddb_database, monkeypatch, capsys
):
    """Test that init of experiment from version with children fails."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    execute(
        "hunt --init-only -n experiment --config orion_config.yaml "
        "./black_box.py -x~normal(0,1)"
    )
    execute(
        "hunt --init-only -n experiment ./black_box.py -x~normal(0,1) -y~+normal(0,1)"
    )

    execute(
        "hunt --init-only -n experiment -v 1 "
        "./black_box.py -x~normal(0,1) -y~+normal(0,1) -z~normal(0,1)",
        assert_code=1,
    )

    captured = capsys.readouterr()
    assert captured.out == ""
    assert "Configuration is different and generates a branching event" in captured.err
    assert "Experiment name" in captured.err


def test_init_w_version_from_exp_wout_child(setup_pickleddb_database, monkeypatch):
    """Test that init of experiment from version without child works."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    execute(
        "hunt --init-only -n experiment --config orion_config.yaml "
        "./black_box.py -x~normal(0,1)"
    )
    execute(
        "hunt --init-only -n experiment ./black_box.py -x~normal(0,1) -y~+normal(0,1)"
    )
    execute(
        "hunt --init-only -n experiment -v 2 ./black_box.py "
        "-x~normal(0,1) -y~+normal(0,1) -z~+normal(0,1)"
    )

    exp = get_storage().fetch_experiments({"name": "experiment", "version": 3})
    assert len(list(exp))


def test_init_w_version_gt_max(setup_pickleddb_database, monkeypatch):
    """Test that init of experiment from version higher than max works."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    execute(
        "hunt --init-only -n experiment --config orion_config.yaml "
        "./black_box.py -x~normal(0,1)"
    )
    execute(
        "hunt --init-only -n experiment ./black_box.py -x~normal(0,1) -y~+normal(0,1)"
    )
    execute(
        "hunt --init-only -n experiment -v 2000 ./black_box.py "
        "-x~normal(0,1) -y~+normal(0,1) -z~+normal(0,1)"
    )

    exp = get_storage().fetch_experiments({"name": "experiment", "version": 3})
    assert len(list(exp))


def test_init_check_increment_w_children(setup_pickleddb_database, monkeypatch):
    """Test that incrementing version works with not same-named children."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    execute(
        "hunt --init-only -n experiment --config orion_config.yaml "
        "./black_box.py -x~normal(0,1)"
    )
    execute(
        "hunt --init-only -n experiment --branch-to experiment_2 ./black_box.py "
        "-x~normal(0,1) -y~+normal(0,1)"
    )
    execute(
        "hunt --init-only -n experiment ./black_box.py -x~normal(0,1) -z~+normal(0,1)"
    )

    exp = get_storage().fetch_experiments({"name": "experiment", "version": 2})
    assert len(list(exp))


def test_branch_from_selected_version(setup_pickleddb_database, monkeypatch):
    """Test that branching from a version passed with `--version` works."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    execute(
        "hunt --init-only -n experiment --config orion_config.yaml "
        "./black_box.py -x~normal(0,1)"
    )
    execute(
        "hunt --init-only -n experiment ./black_box.py -x~normal(0,1) -y~+normal(0,1)"
    )
    execute(
        "hunt --init-only -n experiment --version 1 -b experiment_2 ./black_box.py "
        "-x~normal(0,1) -z~+normal(0,1)"
    )

    storage = get_storage()
    parent = storage.fetch_experiments({"name": "experiment", "version": 1})[0]
    exp = storage.fetch_experiments({"name": "experiment_2"})[0]
    assert exp["refers"]["parent_id"] == parent["_id"]
